from .. import Problem, Mesh
from .. import Utils
from ..Utils import mkvc, matutils, sdiag
from .. import Props
import scipy as sp
import scipy.constants as constants
import os
import numpy as np
import dask
import dask.array as da
from scipy.sparse import csr_matrix as csr
from dask.diagnostics import ProgressBar
import multiprocessing


class GravityIntegral(Problem.LinearProblem):

    rho, rhoMap, rhoDeriv = Props.Invertible("Specific density (g/cc)", default=1.0)

    forwardOnly = False  # If false, matrix is store to memory (watch your RAM)
    actInd = None  #: Active cell indices provided
    verbose = True
    gtgdiag = None
    n_cpu = None
    parallelized = True
    progressIndex = -1
    max_chunk_size = None
    chunk_by_rows = False
    Jpath = "./sensitivity.zarr"
    maxRAM = 8  # Maximum memory usage

    def __init__(self, mesh, **kwargs):
        Problem.BaseProblem.__init__(self, mesh, **kwargs)

        if getattr(self, "actInd", None) is not None:

            if self.actInd.dtype == "bool":
                inds = (
                    np.asarray(
                        [inds for inds, elem in enumerate(self.actInd, 1) if elem],
                        dtype=int,
                    )
                    - 1
                )
            else:
                inds = self.actInd

        else:

            inds = np.asarray(range(self.mesh.nC))

        self.nC = len(inds)

        # Create active cell projector
        P = csr(
            (np.ones(self.nC), (inds, range(self.nC))), shape=(self.mesh.nC, self.nC)
        )

        # Create vectors of nodal location
        # (lower and upper corners for each cell)
        bsw = self.mesh.gridCC - self.mesh.h_gridded / 2.0
        tne = self.mesh.gridCC + self.mesh.h_gridded / 2.0

        xn1, xn2 = bsw[:, 0], tne[:, 0]
        yn1, yn2 = bsw[:, 1], tne[:, 1]

        self.Yn = P.T * np.c_[mkvc(yn1), mkvc(yn2)]
        self.Xn = P.T * np.c_[mkvc(xn1), mkvc(xn2)]

        if self.mesh.dim > 2:
            zn1, zn2 = bsw[:, 2], tne[:, 2]
            self.Zn = P.T * np.c_[mkvc(zn1), mkvc(zn2)]

    def fields(self, m):
        # self.model = self.rhoMap*m

        if self.forwardOnly:

            # Compute the linear operation without forming the full dense G
            return np.array(self.Intrgl_Fwr_Op(m=m), dtype="float")

        else:
            # fields = da.dot(self.G, m)

            return da.dot(
                self.G, (self.rhoMap * m).astype(np.float32)
            )  # np.array(fields, dtype='float')

    def modelMap(self):
        """
            Call for general mapping of the problem
        """
        return self.rhoMap

    def getJtJdiag(self, m, W=None):
        """
            Return the diagonal of JtJ
        """

        dmudm = self.rhoMap.deriv(m)
        self.model = m

        if self.gtgdiag is None:

            if W is None:
                W = np.ones(self.G.shape[1])

            self.gtgdiag = da.sum(
                da.power(W[:, None].astype(np.float32) * self.G, 2), axis=0
            ).compute()

            # for ii in range(self.G.shape[0]):

            #     self.gtgdiag += (w[ii]*self.G[ii, :]*dmudm)**2.

        return mkvc(
            np.sum((sdiag(mkvc(self.gtgdiag) ** 0.5) * dmudm).power(2.0), axis=0)
        )

    def getJ(self, m, f=None):
        """
            Sensitivity matrix
        """

        dmudm = self.rhoMap.deriv(m)

        return da.dot(self.G, dmudm)

    def Jvec(self, m, v, f=None):
        dmudm = self.rhoMap.deriv(m)

        dmudm_v = da.from_array(dmudm * v, chunks=self.G.chunks[1])

        return da.dot(self.G, dmudm_v.astype(np.float32))

    def Jtvec(self, m, v, f=None):

        dmudm = self.rhoMap.deriv(m)
        Jtvec = da.dot(v.astype(np.float32), self.G)
        dmudm_v = dask.delayed(csr.dot)(Jtvec, dmudm)

        return da.from_delayed(dmudm_v, dtype=float, shape=[dmudm.shape[1]])

    @property
    def G(self):
        if not self.ispaired:
            raise Exception("Need to pair!")

        if getattr(self, "_G", None) is None:

            self._G = self.Intrgl_Fwr_Op()

        return self._G

    def Intrgl_Fwr_Op(self, m=None):

        """

        Gravity forward operator in integral form

        flag        = 'z' | 'xyz'

        Return
        _G        = Linear forward modeling operation

        Created on March, 15th 2016

        @author: dominiquef

         """
        if m is not None:
            self.model = self.rhoMap * m

        print("Multiplication")
        self.rxLoc = self.survey.srcField.rxList[0].locs
        self.nD = int(self.rxLoc.shape[0])

        # if self.n_cpu is None:
        #     self.n_cpu = multiprocessing.cpu_count()

        # Switch to determine if the process has to be run in parallel
        job = Forward(
            rxLoc=self.rxLoc,
            Xn=self.Xn,
            Yn=self.Yn,
            Zn=self.Zn,
            n_cpu=self.n_cpu,
            forwardOnly=self.forwardOnly,
            model=self.model,
            components=self.survey.components,
            parallelized=self.parallelized,
            verbose=self.verbose,
            Jpath=self.Jpath,
            maxRAM=self.maxRAM,
            max_chunk_size=self.max_chunk_size,
            chunk_by_rows=self.chunk_by_rows,
        )

        G = job.calculate()

        return G


class Forward:
    """
        Add docstring once it works
    """

    progressIndex = -1
    parallelized = True
    rxLoc = None
    Xn, Yn, Zn = None, None, None
    n_cpu = None
    forwardOnly = False
    model = None
    components = ["gz"]

    verbose = True
    maxRAM = 1
    chunk_by_rows = False

    max_chunk_size = None
    Jpath = "./sensitivity.zarr"

    def __init__(self, **kwargs):
        super().__init__()
        Utils.setKwargs(self, **kwargs)

    def calculate(self):

        self.nD = self.rxLoc.shape[0]
        self.nC = self.Xn.shape[0]

        if self.n_cpu is None:
            self.n_cpu = int(multiprocessing.cpu_count())

        # Set this early so we can get a better memory estimate for dask chunking
        nDataComps = len(self.components)

        if self.parallelized:

            row = dask.delayed(self.calcTrow, pure=True)

            makeRows = [row(self.rxLoc[ii, :]) for ii in range(self.nD)]

            buildMat = [
                da.from_delayed(makeRow, dtype=np.float32, shape=(nDataComps, self.nC))
                for makeRow in makeRows
            ]

            stack = da.vstack(buildMat)

            # Auto rechunk
            # To customise memory use set Dask config in calling scripts: dask.config.set({'array.chunk-size': '128MiB'})
            if self.forwardOnly or self.chunk_by_rows:
                label = "DASK: Chunking by rows"
                # Autochunking by rows is faster and more memory efficient for
                # very large problems sensitivty and forward calculations
                target_size = dask.config.get("array.chunk-size").replace("MiB", " MB")
                stack = stack.rechunk({0: "auto", 1: -1})
            elif self.max_chunk_size:
                label = "DASK: Chunking using parameters"
                # Manual chunking is less sensitive to chunk sizes for some problems
                target_size = f"{self.max_chunk_size:.0f} MB"
                nChunks_col = 1
                nChunks_row = 1
                rowChunk = int(np.ceil(stack.shape[0] / nChunks_row))
                colChunk = int(np.ceil(stack.shape[1] / nChunks_col))
                chunk_size = rowChunk * colChunk * 8 * 1e-6  # in Mb

                # Add more chunks until memory falls below target
                while chunk_size >= self.max_chunk_size:

                    if rowChunk > colChunk:
                        nChunks_row += 1
                    else:
                        nChunks_col += 1

                    rowChunk = int(np.ceil(stack.shape[0] / nChunks_row))
                    colChunk = int(np.ceil(stack.shape[1] / nChunks_col))
                    chunk_size = rowChunk * colChunk * 8 * 1e-6  # in Mb

                stack = stack.rechunk((rowChunk, colChunk))
            else:
                label = "DASK: Chunking by columns"
                # Autochunking by columns is faster for Inversions
                target_size = dask.config.get("array.chunk-size").replace("MiB", " MB")
                stack = stack.rechunk({0: -1, 1: "auto"})

            if self.verbose:
                print(label)
                print("Tile size (nD, nC): ", stack.shape)
                #                print('Chunk sizes (nD, nC): ', stack.chunks) # For debugging only
                print(
                    "Number of chunks: %.0f x %.0f = %.0f"
                    % (
                        len(stack.chunks[0]),
                        len(stack.chunks[1]),
                        len(stack.chunks[0]) * len(stack.chunks[1]),
                    )
                )
                print("Target chunk size: %s" % target_size)
                print(
                    "Max chunk size %.0f x %.0f = %.3f MB"
                    % (
                        max(stack.chunks[0]),
                        max(stack.chunks[1]),
                        max(stack.chunks[0]) * max(stack.chunks[1]) * 8 * 1e-6,
                    )
                )
                print(
                    "Min chunk size %.0f x %.0f = %.3f MB"
                    % (
                        min(stack.chunks[0]),
                        min(stack.chunks[1]),
                        min(stack.chunks[0]) * min(stack.chunks[1]) * 8 * 1e-6,
                    )
                )
                print(
                    "Max RAM (GB x %.0f CPU): %.6f"
                    % (
                        self.n_cpu,
                        max(stack.chunks[0])
                        * max(stack.chunks[1])
                        * 8
                        * 1e-9
                        * self.n_cpu,
                    )
                )
                print(
                    "Tile size (GB): %.3f"
                    % (stack.shape[0] * stack.shape[1] * 8 * 1e-9)
                )

            if self.forwardOnly:

                with ProgressBar():
                    print("Forward calculation: ")
                    pred = da.dot(stack, self.model).compute()

                return pred

            else:

                if os.path.exists(self.Jpath):

                    G = da.from_zarr(self.Jpath)

                    if np.all(
                        np.r_[
                            np.any(np.r_[G.chunks[0]] == stack.chunks[0]),
                            np.any(np.r_[G.chunks[1]] == stack.chunks[1]),
                            np.r_[G.shape] == np.r_[stack.shape],
                        ]
                    ):
                        # Check that loaded G matches supplied data and mesh
                        print(
                            "Zarr file detected with same shape and chunksize ... re-loading"
                        )

                        return G
                    else:

                        print(
                            "Zarr file detected with wrong shape and chunksize ... over-writing"
                        )

                with ProgressBar():
                    print("Saving G to zarr: " + self.Jpath)
                    G = da.to_zarr(
                        stack,
                        self.Jpath,
                        compute=True,
                        return_stored=True,
                        overwrite=True,
                    )

        else:

            result = []
            for ii in range(self.nD):
                result += [self.calcTrow(self.rxLoc[ii, :])]
                self.progress(ii, self.nD)

            G = np.vstack(result)

        return G

    def calcTrow(self, receiver_location):
        """
            Load in the active nodes of a tensor mesh and computes the magnetic
            forward relation between a cuboid and a given observation
            location outside the Earth [obsx, obsy, obsz]

            INPUT:
            xyzLoc:  [obsx, obsy, obsz] nC x 3 Array

            OUTPUT:
            Tx = [Txx Txy Txz]
            Ty = [Tyx Tyy Tyz]
            Tz = [Tzx Tzy Tzz]

        """
        eps = 1e-8

        NewtG = constants.G * 1e8

        dx = self.Xn - receiver_location[0]
        dy = self.Yn - receiver_location[1]
        dz = self.Zn - receiver_location[2]

        compDict = {key: np.zeros(self.Xn.shape[0]) for key in self.components}

        gxx = np.zeros(self.Xn.shape[0])
        gyy = np.zeros(self.Xn.shape[0])

        for aa in range(2):
            for bb in range(2):
                for cc in range(2):

                    r = (
                        mkvc(dx[:, aa]) ** 2
                        + mkvc(dy[:, bb]) ** 2
                        + mkvc(dz[:, cc]) ** 2
                    ) ** (0.50) + eps

                    dz_r = dz[:, cc] + r + eps
                    dy_r = dy[:, bb] + r + eps
                    dx_r = dx[:, aa] + r + eps

                    dxr = dx[:, aa] * r + eps
                    dyr = dy[:, bb] * r + eps
                    dzr = dz[:, cc] * r + eps

                    dydz = dy[:, bb] * dz[:, cc]
                    dxdy = dx[:, aa] * dy[:, bb]
                    dxdz = dx[:, aa] * dz[:, cc]

                    if "gx" in self.components:
                        compDict["gx"] += (
                            (-1) ** aa
                            * (-1) ** bb
                            * (-1) ** cc
                            * (
                                dy[:, bb] * np.log(dz_r)
                                + dz[:, cc] * np.log(dy_r)
                                - dx[:, aa] * np.arctan(dydz / dxr)
                            )
                        )

                    if "gy" in self.components:
                        compDict["gy"] += (
                            (-1) ** aa
                            * (-1) ** bb
                            * (-1) ** cc
                            * (
                                dx[:, aa] * np.log(dz_r)
                                + dz[:, cc] * np.log(dx_r)
                                - dy[:, bb] * np.arctan(dxdz / dyr)
                            )
                        )

                    if "gz" in self.components:
                        compDict["gz"] += (
                            (-1) ** aa
                            * (-1) ** bb
                            * (-1) ** cc
                            * (
                                dx[:, aa] * np.log(dy_r)
                                + dy[:, bb] * np.log(dx_r)
                                - dz[:, cc] * np.arctan(dxdy / dzr)
                            )
                        )

                    arg = dy[:, bb] * dz[:, cc] / dxr

                    if (
                        ("gxx" in self.components)
                        or ("gzz" in self.components)
                        or ("guv" in self.components)
                    ):
                        gxx -= (
                            1e4
                            * (-1) ** aa
                            * (-1) ** bb
                            * (-1) ** cc
                            * (
                                dxdy / (r * dz_r + eps)
                                + dxdz / (r * dy_r + eps)
                                - np.arctan(arg + eps)
                                + dx[:, aa]
                                * (1.0 / (1 + arg ** 2.0))
                                * dydz
                                / dxr ** 2.0
                                * (r + dx[:, aa] ** 2.0 / r)
                            )
                        )

                    if "gxy" in self.components:
                        compDict["gxy"] -= (
                            1e4
                            * (-1) ** aa
                            * (-1) ** bb
                            * (-1) ** cc
                            * (
                                np.log(dz_r)
                                + dy[:, bb] ** 2.0 / (r * dz_r)
                                + dz[:, cc] / r
                                - 1.0
                                / (1 + arg ** 2.0 + eps)
                                * (dz[:, cc] / r ** 2)
                                * (r - dy[:, bb] ** 2.0 / r)
                            )
                        )

                    if "gxz" in self.components:
                        compDict["gxz"] -= (
                            1e4
                            * (-1) ** aa
                            * (-1) ** bb
                            * (-1) ** cc
                            * (
                                np.log(dy_r)
                                + dz[:, cc] ** 2.0 / (r * dy_r)
                                + dy[:, bb] / r
                                - 1.0
                                / (1 + arg ** 2.0)
                                * (dy[:, bb] / (r ** 2))
                                * (r - dz[:, cc] ** 2.0 / r)
                            )
                        )

                    arg = dx[:, aa] * dz[:, cc] / dyr

                    if (
                        ("gyy" in self.components)
                        or ("gzz" in self.components)
                        or ("guv" in self.components)
                    ):
                        gyy -= (
                            1e4
                            * (-1) ** aa
                            * (-1) ** bb
                            * (-1) ** cc
                            * (
                                dxdy / (r * dz_r + eps)
                                + dydz / (r * dx_r + eps)
                                - np.arctan(arg + eps)
                                + dy[:, bb]
                                * (1.0 / (1 + arg ** 2.0 + eps))
                                * dxdz
                                / dyr ** 2.0
                                * (r + dy[:, bb] ** 2.0 / r)
                            )
                        )

                    if "gyz" in self.components:
                        compDict["gyz"] -= (
                            1e4
                            * (-1) ** aa
                            * (-1) ** bb
                            * (-1) ** cc
                            * (
                                np.log(dx_r)
                                + dz[:, cc] ** 2.0 / (r * (dx_r))
                                + dx[:, aa] / r
                                - 1.0
                                / (1 + arg ** 2.0)
                                * (dx[:, aa] / (r ** 2))
                                * (r - dz[:, cc] ** 2.0 / r)
                            )
                        )

        if "gyy" in self.components:
            compDict["gyy"] = gyy

        if "gxx" in self.components:
            compDict["gxx"] = gxx

        if "gzz" in self.components:
            compDict["gzz"] = -gxx - gyy

        if "guv" in self.components:
            compDict["guv"] = -0.5 * (gxx - gyy)

        return np.vstack([NewtG * compDict[key] for key in list(compDict.keys())])


class Problem3D_Diff(Problem.BaseProblem):
    """
        Gravity in differential equations!
    """

    _depreciate_main_map = "rhoMap"

    rho, rhoMap, rhoDeriv = Props.Invertible("Specific density (g/cc)", default=1.0)

    solver = None

    def __init__(self, mesh, **kwargs):
        Problem.BaseProblem.__init__(self, mesh, **kwargs)

        self.mesh.setCellGradBC("dirichlet")

        self._Div = self.mesh.cellGrad

    @property
    def MfI(self):
        return self._MfI

    @property
    def Mfi(self):
        return self._Mfi

    def makeMassMatrices(self, m):
        self.model = m
        self._Mfi = self.mesh.getFaceInnerProduct()
        self._MfI = Utils.sdiag(1.0 / self._Mfi.diagonal())

    def getRHS(self, m):
        """


        """

        Mc = Utils.sdiag(self.mesh.vol)

        self.model = m
        rho = self.rho

        return Mc * rho

    def getA(self, m):
        r"""
        GetA creates and returns the A matrix for the Gravity nodal problem

        The A matrix has the form:

        .. math ::

            \mathbf{A} =  \Div(\MfMui)^{-1}\Div^{T}

        """
        return -self._Div.T * self.Mfi * self._Div

    def fields(self, m):
        r"""
            Return magnetic potential (u) and flux (B)
            u: defined on the cell nodes [nC x 1]
            gField: defined on the cell faces [nF x 1]

            After we compute u, then we update B.

            .. math ::

                \mathbf{B}_s = (\MfMui)^{-1}\mathbf{M}^f_{\mu_0^{-1}}\mathbf{B}_0-\mathbf{B}_0 -(\MfMui)^{-1}\Div^T \mathbf{u}

        """
        from scipy.constants import G as NewtG

        self.makeMassMatrices(m)
        A = self.getA(m)
        RHS = self.getRHS(m)

        if self.solver is None:
            m1 = sp.linalg.interface.aslinearoperator(Utils.sdiag(1 / A.diagonal()))
            u, info = sp.linalg.bicgstab(A, RHS, tol=1e-6, maxiter=1000, M=m1)

        else:
            print("Solving with Paradiso")
            Ainv = self.solver(A)
            u = Ainv * RHS

        gField = 4.0 * np.pi * NewtG * 1e8 * self._Div * u

        return {"G": gField, "u": u}



import SimPEG
from SimPEG.Utils import Zero, closestPoints, mkvc
import numpy as np


class BaseSrc(SimPEG.Survey.BaseSrc):

    current = 1.0
    loc = None

    def __init__(self, rxList, **kwargs):
        SimPEG.Survey.BaseSrc.__init__(self, rxList, **kwargs)

    def eval(self, prob):
        raise NotImplementedError

    def evalDeriv(self, prob):
        return Zero()

    @property
    def nD(self):
        """Number of data"""
        return self.vnD.sum()

    @property
    def vnD(self):
        """Vector number of data"""
        return np.array([rx.nD * len(rx.times) for rx in self.rxList])


class Dipole(BaseSrc):
    def __init__(self, rxList, locA, locB, **kwargs):
        assert locA.shape == locB.shape, "Shape of locA and locB should be the same"
        self.loc = [locA, locB]
        BaseSrc.__init__(self, rxList, **kwargs)

    def eval(self, prob):
        if prob._formulation == "HJ":
            inds = closestPoints(prob.mesh, self.loc, gridLoc="CC")
            q = np.zeros(prob.mesh.nC)
            q[inds] = self.current * np.r_[1.0, -1.0]
        elif prob._formulation == "EB":
            qa = prob.mesh.getInterpolationMat(self.loc[0], locType="N").todense()
            qb = -prob.mesh.getInterpolationMat(self.loc[1], locType="N").todense()
            q = self.current * mkvc(qa + qb)
        return q


class Pole(BaseSrc):
    def __init__(self, rxList, loc, **kwargs):
        BaseSrc.__init__(self, rxList, loc=loc, **kwargs)

    def eval(self, prob):
        if prob._formulation == "HJ":
            inds = closestPoints(prob.mesh, self.loc)
            q = np.zeros(prob.mesh.nC)
            q[inds] = self.current * np.r_[1.0]
        elif prob._formulation == "EB":
            q = prob.mesh.getInterpolationMat(self.loc, locType="N").todense()
            q = self.current * mkvc(q)
        return q



from discretize.utils.interputils import interpmat

from .matutils import (
    mkvc,
    sdiag,
    sdInv,
    speye,
    kron3,
    spzeros,
    ddx,
    av,
    av_extrap,
    ndgrid,
    ind2sub,
    sub2ind,
    getSubArray,
    inv3X3BlockDiagonal,
    inv2X2BlockDiagonal,
    TensorType,
    makePropertyTensor,
    invPropertyTensor,
    diagEst,
    Zero,
    Identity,
    uniqueRows,
)
from .codeutils import (
    memProfileWrapper,
    hook,
    setKwargs,
    printTitles,
    printLine,
    checkStoppers,
    printStoppers,
    callHooks,
    dependentProperty,
    asArray_N_x_Dim,
    requires,
)
from .meshutils import exampleLrmGrid, meshTensor, closestPoints, ExtractCoreMesh
from .curvutils import volTetra, faceInfo, indexCube
from .CounterUtils import Counter, count, timeIt
from . import ModelBuilder
from . import SolverUtils
from .coordutils import rotatePointsFromNormals, rotationMatrixFromNormals
from .modelutils import surface2ind_topo
from .PlotUtils import plot2Ddata, plotLayer
from .io_utils import download

from .printinfo import versions



import json
import os

import ipywidgets as widgets
import matplotlib.pyplot as plt
import numpy as np
from geoh5py.groups import ContainerGroup
from geoh5py.objects import BlockModel, Curve, Grid2D, Octree, Points, Surface
from geoh5py.workspace import Workspace
from ipywidgets.widgets import (
    Dropdown,
    HBox,
    Label,
    Layout,
    Text,
    VBox,
    Checkbox,
    FloatText,
    ToggleButton,
    IntText,
    Widget,
)

from geoapps.base import BaseApplication, working_copy
from geoapps.plotting import plot_plan_data_selection, PlotSelection2D
from geoapps.utils import find_value, rotate_xy, geophysical_systems
from geoapps.selection import ObjectDataSelection, LineOptions


class ChannelOptions(BaseApplication):
    """
    Options for data channels
    """

    def __init__(self, key, description, **kwargs):

        self._active = Checkbox(value=False, indent=True, description="Active",)
        self._label = widgets.Text(description=description)
        self._channel_selection = Dropdown(description="Associated Data:")
        self._channel_selection.header = key

        self._uncertainties = widgets.Text(description="Error (%, floor)")
        self._offsets = widgets.Text(description="Offsets (x,y,z)")

        self._widget = VBox(
            [
                self._active,
                self._label,
                self._channel_selection,
                self._uncertainties,
                self._offsets,
            ]
        )

        super().__init__(**kwargs)

    @property
    def active(self):
        return self._active

    @property
    def label(self):
        return self._label

    @property
    def channel_selection(self):
        return self._channel_selection

    @property
    def uncertainties(self):
        return self._uncertainties

    @property
    def offsets(self):
        return self._offsets

    @property
    def widget(self):
        return self._widget


class SensorOptions(BaseApplication):
    """
    Define the receiver spatial parameters
    """

    def __init__(self, **kwargs):

        super().__init__(**kwargs)

        self.selection = ObjectDataSelection(**kwargs)
        self._objects = self.selection.objects
        self._value = self.selection.data

        self._value.description = "Height Channel"

        self._offset = Text(description="[dx, dy, dz]", value="0, 0, 0")

        self._constant = FloatText(description="Constant elevation (m)",)
        if "offset" in kwargs.keys():
            self._offset.value = kwargs["value"]

        self._options = {
            "(x, y, z) + offset(x,y,z)": self._offset,
            "(x, y, topo + radar) + offset(x,y,z)": VBox([self._offset, self._value]),
        }
        self._options_button = widgets.RadioButtons(
            options=[
                "(x, y, z) + offset(x,y,z)",
                "(x, y, topo + radar) + offset(x,y,z)",
            ],
            description="Define by:",
        )

        def update_options(_):
            self.update_options()

        self._options_button.observe(update_options)
        self._widget = VBox(
            [self._options_button, self._options[self._options_button.value]]
        )

        super().__init__(**kwargs)

        def update_list(_):
            self.update_list()

        self._objects.observe(update_list)

        self.update_list()

        if "value" in kwargs.keys() and kwargs["value"] in self._value.options:
            self._value.value = kwargs["value"]

    @property
    def objects(self):
        return self._objects

    @property
    def offset(self):
        return self._offset

    @property
    def options(self):
        return self._options

    @property
    def options_button(self):
        return self._options_button

    def update_list(self):
        if self.workspace.get_entity(self.objects.value):
            obj = self.workspace.get_entity(self.objects.value)[0]
            data_list = obj.get_data_list() + [""]

            self._value.options = [
                name for name in data_list if "visual" not in name.lower()
            ]

            self._value.value = ""

    def update_options(self):
        self._widget.children = [
            self._options_button,
            self._options[self._options_button.value],
        ]

    @property
    def value(self):
        return self._value

    @property
    def widget(self):
        return self._widget


class TopographyOptions(BaseApplication):
    """
    Define the topography used by the inversion
    """

    def __init__(self, **kwargs):

        self.selection = ObjectDataSelection(**kwargs)
        self._objects = self.selection.objects
        self._value = self.selection.data
        self._offset = FloatText(description="Vertical offset (m)")
        self._constant = FloatText(description="Constant elevation (m)",)

        def update_list(_):
            self.update_list()

        self._objects.observe(update_list, names="value")

        super().__init__(**kwargs)

        self.objects.value = find_value(self.objects.options, ["topo", "dem", "dtm"])

        self._options = {
            "Object": self.selection.widget,
            "Drape Height": self.offset,
            "Constant": self.constant,
            "None": widgets.Label("No topography"),
        }
        self._options_button = widgets.RadioButtons(
            options=["Object", "Drape Height", "Constant"], description="Define by:",
        )

        def update_options(_):
            self.update_options()

        self._options_button.observe(update_options)
        self._widget = VBox(
            [self._options_button, self._options[self._options_button.value]]
        )

    @property
    def panel(self):
        return self._panel

    @property
    def constant(self):
        return self._constant

    @property
    def objects(self):
        return self._objects

    @property
    def offset(self):
        return self._offset

    @property
    def options(self):
        return self._options

    @property
    def options_button(self):
        return self._options_button

    def update_list(self):
        w_s = Workspace(self.h5file)
        if w_s.get_entity(self.objects.value):
            data_list = w_s.get_entity(self.objects.value)[0].get_data_list()
            self._value.options = [
                name for name in data_list if "visual" not in name.lower()
            ] + ["Vertices"]
            self._value.value = find_value(
                self._value.options, ["dem", "topo", "dtm"], default="Vertices"
            )

    def update_options(self):
        self._widget.children = [
            self._options_button,
            self._options[self._options_button.value],
        ]

    @property
    def value(self):
        return self._value

    @property
    def widget(self):
        return self._widget


class MeshOctreeOptions(BaseApplication):
    """
    Widget used for the creation of an octree meshes
    """

    def __init__(self, **kwargs):
        self._core_cell_size = widgets.Text(
            value="25, 25, 25", description="Smallest cells",
        )
        self._octree_levels_topo = widgets.Text(
            value="0, 0, 0, 2", description="Layers below topo",
        )
        self._octree_levels_obs = widgets.Text(
            value="5, 5, 5, 5", description="Layers below data",
        )
        self._depth_core = FloatText(value=500, description="Minimum depth (m)",)
        self._padding_distance = widgets.Text(
            value="0, 0, 0, 0, 0, 0", description="Padding [W,E,N,S,D,U] (m)",
        )

        self._max_distance = FloatText(
            value=1000, description="Max triangulation length",
        )

        self._widget = widgets.VBox(
            [
                Label("Octree Mesh"),
                self._core_cell_size,
                self._octree_levels_topo,
                self._octree_levels_obs,
                self._depth_core,
                self._padding_distance,
                self._max_distance,
            ]
        )

        super().__init__(**kwargs)

    @property
    def core_cell_size(self):
        return self._core_cell_size

    @property
    def depth_core(self):
        return self._depth_core

    @property
    def max_distance(self):
        return self._max_distance

    @property
    def octree_levels_obs(self):
        return self._octree_levels_obs

    @property
    def octree_levels_topo(self):
        return self._octree_levels_topo

    @property
    def padding_distance(self):
        return self._padding_distance

    @property
    def widget(self):
        return self._widget


class Mesh1DOptions(BaseApplication):
    """
    Widget used for the creation of a 1D mesh
    """

    def __init__(self, **kwargs):
        self._hz_expansion = FloatText(value=1.05, description="Expansion factor:",)
        self._hz_min = FloatText(value=10.0, description="Smallest cell (m):",)
        self._n_cells = FloatText(value=25.0, description="Number of cells:",)

        super().__init__(**kwargs)

        self.cell_count = Label(f"Max depth: {self.count_cells():.2f} m")

        def update_hz_count(_):
            self.update_hz_count()

        self.n_cells.observe(update_hz_count)
        self.hz_expansion.observe(update_hz_count)
        self.hz_min.observe(update_hz_count)

        self._widget = VBox(
            [
                Label("1D Mesh"),
                self.hz_min,
                self.hz_expansion,
                self.n_cells,
                self.cell_count,
            ]
        )

    def count_cells(self):
        return (
            self.hz_min.value * self.hz_expansion.value ** np.arange(self.n_cells.value)
        ).sum()

    def update_hz_count(self):
        self.cell_count.value = f"Max depth: {self.count_cells():.2f} m"

    @property
    def hz_expansion(self):
        """

        """
        return self._hz_expansion

    @property
    def hz_min(self):
        """

        """
        return self._hz_min

    @property
    def n_cells(self):
        """

        """
        return self._n_cells

    @property
    def widget(self):
        return self._widget


class ModelOptions(BaseApplication):
    """
    Widgets for the selection of model options
    """

    def __init__(self, name_list, **kwargs):

        self._options = widgets.RadioButtons(
            options=["Model", "Value"], value="Value", disabled=False,
        )

        def update_panel(_):
            self.update_panel()

        self._options.observe(update_panel, names="value")
        self._list = widgets.Dropdown(description="3D Model", options=name_list)
        self._value = FloatText(description="Units")
        self._description = Label()
        self._widget = widgets.VBox(
            [self._description, widgets.VBox([self._options, self._value])]
        )

        super().__init__(**kwargs)

        for key, value in kwargs.items():
            if key == "units":
                self._value.description = value
            elif key == "label":
                self._description.value = value

    def update_panel(self):

        if self._options.value == "Model":
            self._widget.children[1].children = [self._options, self._list]
            self._widget.children[1].children[1].layout.visibility = "visible"
        elif self._options.value == "Value":
            self._widget.children[1].children = [self._options, self._value]
            self._widget.children[1].children[1].layout.visibility = "visible"
        else:
            self._widget.children[1].children[1].layout.visibility = "hidden"

    @property
    def description(self):
        return self._description

    @property
    def list(self):
        return self._list

    @property
    def options(self):
        return self._options

    @property
    def value(self):
        return self._value

    @property
    def widget(self):
        return self._widget


class InversionOptions(BaseApplication):
    """
    Collection of widgets controlling the inversion parameters
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        model_list = []
        for obj in self.workspace.all_objects():
            if isinstance(obj, (BlockModel, Octree, Surface)):
                for data in obj.children:
                    if (
                        getattr(data, "values", None) is not None
                        and data.name != "Visual Parameters"
                    ):
                        model_list += [data.name]

        self._output_name = widgets.Text(
            value="Inversion_", description="Save to:", disabled=False
        )

        self._chi_factor = FloatText(
            value=1, description="Target misfit", disabled=False
        )
        self._uncert_mode = widgets.RadioButtons(
            options=[
                "Estimated (%|data| + background)",
                r"User input (\%|data| + floor)",
            ],
            value=r"User input (\%|data| + floor)",
            disabled=False,
        )
        self._lower_bound = widgets.Text(value=None, description="Lower bound value",)
        self._upper_bound = widgets.Text(value=None, description="Upper bound value",)

        self._ignore_values = widgets.Text(value="<0", tooltip="Dummy value",)
        self._max_iterations = IntText(value=10, description="Max beta Iterations")
        self._max_cg_iterations = IntText(value=10, description="Max CG Iterations")
        self._tol_cg = FloatText(value=1e-4, description="CG Tolerance")

        self._beta_start_options = widgets.RadioButtons(
            options=["value", "ratio"],
            value="ratio",
            description="Starting tradeoff (beta):",
        )
        self._beta_start = FloatText(value=1e2, description="phi_d/phi_m")

        def initial_beta_change(_):
            self.initial_beta_change()

        self._beta_start_options.observe(initial_beta_change)

        self._beta_start_panel = HBox([self._beta_start_options, self._beta_start])
        self._optimization = VBox(
            [
                self._max_iterations,
                self._chi_factor,
                self._beta_start_panel,
                self._max_cg_iterations,
                self._tol_cg,
            ]
        )
        self._starting_model = ModelOptions(model_list)
        self._susceptibility_model = ModelOptions(
            model_list, units="SI", label="Background susceptibility"
        )
        self._susceptibility_model.options.options = ["None", "Model", "Value"]
        self._reference_model = ModelOptions(model_list)
        self._reference_model.options.options = [
            "None",
            "Best-fitting halfspace",
            "Model",
            "Value",
        ]

        def update_ref(_):
            self.update_ref()

        self.reference_model.options.observe(update_ref)

        self._alphas = widgets.Text(
            value="1, 1, 1, 1", description="Scaling alpha_(s, x, y, z)",
        )
        self._norms = widgets.Text(
            value="2, 2, 2, 2", description="Norms p_(s, x, y, z)",
        )

        def check_max_iterations(_):
            self.check_max_iterations()

        self._norms.observe(check_max_iterations)

        self._mesh = MeshOctreeOptions()
        self.inversion_options = {
            "output name": self._output_name,
            "uncertainties": self._uncert_mode,
            "starting model": self._starting_model.widget,
            "background susceptibility": self._susceptibility_model.widget,
            "regularization": VBox(
                [self._reference_model.widget, self._alphas, self._norms]
            ),
            "upper-lower bounds": VBox([self._upper_bound, self._lower_bound]),
            "mesh": self.mesh.widget,
            "ignore values (<0 = no negatives)": self._ignore_values,
            "optimization": self._optimization,
        }
        self.option_choices = widgets.Dropdown(
            options=list(self.inversion_options.keys()),
            value=list(self.inversion_options.keys())[0],
            disabled=False,
        )

        def inversion_option_change(_):
            self.inversion_option_change()

        self.option_choices.observe(inversion_option_change)

        self._widget = widgets.VBox(
            [
                widgets.HBox([widgets.Label("Inversion Options")]),
                widgets.HBox(
                    [
                        self.option_choices,
                        self.inversion_options[self.option_choices.value],
                    ],
                ),
            ],
            layout=Layout(width="100%"),
        )

        super().__init__(**kwargs)

    def check_max_iterations(self):
        if not all([val == 2 for val in string_2_list(self._norms.value)]):
            self._max_iterations.value = 30
        else:
            self._max_iterations.value = 10

    def inversion_option_change(self):
        self._widget.children[1].children = [
            self.option_choices,
            self.inversion_options[self.option_choices.value],
        ]

    def initial_beta_change(self):
        if self._beta_start_options.value == "ratio":
            self._beta_start.description = "phi_d/phi_m"
        else:
            self._beta_start.description = ""

    @property
    def alphas(self):
        return self._alphas

    @property
    def beta_start(self):
        return self._beta_start

    @property
    def beta_start_options(self):
        return self._beta_start_options

    @property
    def chi_factor(self):
        return self._chi_factor

    @property
    def ignore_values(self):
        return self._ignore_values

    @property
    def lower_bound(self):
        return self._lower_bound

    @property
    def upper_bound(self):
        return self._upper_bound

    @property
    def max_iterations(self):
        return self._max_iterations

    @property
    def max_cg_iterations(self):
        return self._max_cg_iterations

    @property
    def mesh(self):
        return self._mesh

    @property
    def norms(self):
        return self._norms

    @property
    def output_name(self):
        return self._output_name

    @property
    def reference_model(self):
        return self._reference_model

    @property
    def starting_model(self):
        return self._starting_model

    @property
    def susceptibility_model(self):
        return self._susceptibility_model

    @property
    def tol_cg(self):
        return self._tol_cg

    @property
    def uncert_mode(self):
        return self._uncert_mode

    @property
    def widget(self):
        """
        :obj:`ipywidgets.VBox`: Pre-defined application layout
        """
        return self._widget

    def update_ref(self):
        alphas = string_2_list(self.alphas.value)
        if self.reference_model.options.value == "None":
            alphas[0] = 0.0
        else:
            alphas[0] = 1.0
        self.alphas.value = ", ".join(list(map(str, alphas)))


def get_inversion_output(h5file, group_name):
    """
    Recover an inversion iterations from a ContainerGroup comments.
    """
    workspace = Workspace(h5file)
    out = {"time": [], "iteration": [], "phi_d": [], "phi_m": [], "beta": []}

    if workspace.get_entity(group_name):
        group = workspace.get_entity(group_name)[0]

        for comment in group.comments.values:
            if "Iteration" in comment["Author"]:
                out["iteration"] += [np.int(comment["Author"].split("_")[1])]
                out["time"] += [comment["Date"]]

                values = json.loads(comment["Text"])

                out["phi_d"] += [np.float(values["phi_d"])]
                out["phi_m"] += [np.float(values["phi_m"])]
                out["beta"] += [np.float(values["beta"])]

        if len(out["iteration"]) > 0:
            out["iteration"] = np.hstack(out["iteration"])
            ind = np.argsort(out["iteration"])
            out["iteration"] = out["iteration"][ind]
            out["phi_d"] = np.hstack(out["phi_d"])[ind]
            out["phi_m"] = np.hstack(out["phi_m"])[ind]
            out["time"] = np.hstack(out["time"])[ind]

    return out


def plot_convergence_curve(h5file):
    """

    """
    workspace = Workspace(h5file)

    names = [
        group.name
        for group in workspace.all_groups()
        if isinstance(group, ContainerGroup)
    ]

    objects = widgets.Dropdown(
        options=names, value=names[0], description="Inversion Group:",
    )

    def plot_curve(objects):

        inversion = workspace.get_entity(objects)[0]
        result = None
        if getattr(inversion, "comments", None) is not None:
            if inversion.comments.values is not None:
                result = get_inversion_output(workspace.h5file, objects)
                iterations = result["iteration"]
                phi_d = result["phi_d"]
                phi_m = result["phi_m"]

                ax1 = plt.subplot()
                ax2 = ax1.twinx()
                ax1.plot(iterations, phi_d, linewidth=3, c="k")
                ax1.set_xlabel("Iterations")
                ax1.set_ylabel(r"$\phi_d$", size=16)
                ax2.plot(iterations, phi_m, linewidth=3, c="r")
                ax2.set_ylabel(r"$\phi_m$", size=16)

        return result

    interactive_plot = widgets.interactive(plot_curve, objects=objects)

    return interactive_plot


def inversion_defaults():
    """
    Get defaults for gravity, magnetics and EM1D inversions
    """
    defaults = {
        "units": {"Gravity": "g/cc", "Magnetics": "SI", "EM1D": "S/m"},
        "property": {
            "Gravity": "density",
            "Magnetics": "effective susceptibility",
            "EM1D": "conductivity",
        },
        "reference_value": {"Gravity": 0.0, "Magnetics": 0.0, "EM1D": 1e-3},
        "starting_value": {"Gravity": 1e-4, "Magnetics": 1e-4, "EM1D": 1e-3},
    }

    return defaults


def string_2_list(string):
    """
    Convert a list of numbers separated by comma to a list of floats
    """
    return [np.float(val) for val in string.split(",") if len(val) > 0]


class InversionApp(BaseApplication):
    def __init__(self, **kwargs):

        kwargs = working_copy(**kwargs)

        super().__init__(**kwargs)

        # Load all known em systems
        self.em_system_specs = geophysical_systems.parameters()
        self._data_count = (Label("Data Count: 0", tooltip="Keep <1500 for speed"),)
        self._forward_only = Checkbox(
            value=False,
            description="Forward only",
            tooltip="Forward response of reference model",
        )
        self._inducing_field = widgets.Text(
            description="Inducing Field [Amp, Inc, Dec]",
        )
        self._inversion_parameters = InversionOptions(**kwargs)
        self.mesh_octree = MeshOctreeOptions(**kwargs)
        self.mesh_1D = Mesh1DOptions(**kwargs)
        self._run = ToggleButton(
            value=False, description="Run SimPEG", button_style="danger", icon="check"
        )
        self._starting_channel = (IntText(value=None, description="Starting Channel"),)
        self._system = Dropdown(
            options=["Magnetics", "Gravity"] + list(self.em_system_specs.keys()),
            description="Survey Type: ",
        )
        self._write = ToggleButton(
            value=False,
            description="Write input",
            button_style="warning",
            tooltip="Write json input file",
            icon="check",
        )
        self.plot_selection = PlotSelection2D(
            select_multiple=True, add_groups=True, **kwargs
        )
        self.selection = self.plot_selection.selection

        # Link the data extent with the octree mesh param
        def update_octree_param(_):
            self.update_octree_param()

        for item in self.plot_selection.__dict__.values():
            if isinstance(item, Widget):
                item.observe(update_octree_param)

        dsep = os.path.sep
        if self.h5file is not None:
            self._inv_dir = (
                dsep.join(os.path.dirname(os.path.abspath(self.h5file)).split(dsep))
                + dsep
            )
        else:
            self._inv_dir = os.getcwd() + dsep

        def run_trigger(_):
            self.run_trigger()

        self.run.observe(run_trigger)

        def write_trigger(_):
            self.write_trigger()

        self.write.observe(write_trigger)

        def system_observer(_):
            self.system_observer()

        self.system.observe(system_observer, names="value")

        def update_options(_):
            self.update_options()

        for item in self.inversion_parameters.__dict__.values():
            if isinstance(item, Widget):
                item.observe(update_options)
            elif isinstance(item, BaseApplication):
                for val in item.__dict__.values():
                    if isinstance(val, Widget):
                        val.observe(update_options)

        super().__init__(**kwargs)

        self.plot_selection.widget.layout = Layout(width="60%")

        def object_observer(_):
            self.object_observer()

        self.selection.objects.observe(object_observer, names="value")

        def update_component_panel(_):
            self.update_component_panel()

        self.selection.data.observe(update_component_panel, names="value")

        self.data_channel_choices = widgets.Dropdown()

        def data_channel_choices_observer(_):
            self.data_channel_choices_observer()

        self.data_channel_choices.observe(data_channel_choices_observer, names="value")
        self.data_channel_panel = widgets.VBox([self.data_channel_choices])
        self.survey_type_panel = VBox([self.system])

        self.topography = TopographyOptions(**kwargs)

        # Define bird parameters
        sub_kwargs = kwargs.copy()
        if "objects" in sub_kwargs.keys():
            del sub_kwargs["objects"]

        self.sensor = SensorOptions(objects=self.selection.objects, **sub_kwargs)
        self.lines = LineOptions(
            objects=self.selection.objects,
            select_multiple_lines=True,
            find_value="line",
            **sub_kwargs,
        )

        def update_selection(_):
            self.update_selection()

        self.lines.lines.observe(update_selection, names="value")

        # SPATIAL PARAMETERS DROPDOWN
        self.spatial_options = {
            "Topography": self.topography.widget,
            "Receivers": self.sensor.widget,
            "Line ID": self.lines.widget,
        }
        self.spatial_choices = widgets.Dropdown(
            options=list(self.spatial_options.keys()),
            value=list(self.spatial_options.keys())[0],
            disabled=False,
        )

        def spatial_option_change(_):
            self.spatial_option_change()

        self.spatial_choices.observe(spatial_option_change)

        self.spatial_panel = VBox(
            [self.spatial_choices, self.spatial_options[self.spatial_choices.value]]
        )

        self._widget = VBox(
            [
                HBox(
                    [
                        self.plot_selection.widget,
                        VBox(
                            [
                                VBox([Label(""), self.survey_type_panel]),
                                VBox([Label("Components"), self.data_channel_panel]),
                            ],
                            layout=Layout(width="40%"),
                        ),
                    ]
                ),
                VBox([Label("Spatial Information"), self.spatial_panel]),
                self.inversion_parameters.widget,
                self.forward_only,
                self.write,
                self.run,
            ]
        )

        self.object_observer()
        self.inversion_parameters.update_ref()

    @property
    def data_count(self):
        """

        """
        return self._data_count

    @property
    def forward_only(self):
        """

        """
        return self._forward_only

    @property
    def inducing_field(self):
        """

        """
        return self._inducing_field

    @property
    def inversion_parameters(self):
        """

        """
        return self._inversion_parameters

    @property
    def run(self):
        """

        """
        return self._run

    @property
    def starting_channel(self):
        """

        """
        return self._starting_channel

    @property
    def system(self):
        """

        """
        return self._system

    @property
    def widget(self):
        """
        :obj:`ipywidgets.VBox`: Pre-defined application layout
        """
        return self._widget

    @property
    def write(self):
        """

        """
        return self._write

    def run_trigger(self):
        """

        """
        if self.run.value:
            if self.system.value in ["Gravity", "Magnetics"]:
                os.system(
                    "start cmd.exe @cmd /k "
                    + 'python -m geoapps.pf_inversion "'
                    + self._inv_dir
                    + f'\\{self.inversion_parameters.output_name.value}.json"'
                )
            else:
                os.system(
                    "start cmd.exe @cmd /k "
                    + 'python -m geoapps.em1d_inversion "'
                    + self._inv_dir
                    + f'\\{self.inversion_parameters.output_name.value}.json"'
                )
            self.run.value = False
            self.run.button_style = ""

    # Check parameters change
    def update_options(self):
        self.write.button_style = "warning"
        self.run.button_style = "danger"

    def system_observer(self):
        """
        Change the application on change of system
        """
        if self.system.value in ["Magnetics", "Gravity"]:
            start_channel = 0
            if self.system.value == "Magnetics":
                data_type_list = ["tmi", "bxx", "bxy", "bxz", "byy", "byz", "bzz"]
                labels = ["tmi", "bxx", "bxy", "bxz", "byy", "byz", "bzz"]

            else:
                data_type_list = ["gz", "gxx", "gxy", "gxz", "gyy", "gyz", "gzz"]
                labels = ["gz", "gxx", "gxy", "gxz", "gyy", "gyz", "gzz"]

            tx_offsets = [[0, 0, 0]]
            uncertainties = [[0, 1]] * len(data_type_list)

            system_specs = {}
            for key in data_type_list:
                system_specs[key] = key

            # Remove line_id from choices
            self.spatial_choices.options = list(self.spatial_options.keys())[:2]

            # Switch mesh options
            self.inversion_parameters._mesh = self.mesh_octree
            self.inversion_parameters.inversion_options[
                "mesh"
            ] = self.mesh_octree.widget

            self.inversion_parameters.reference_model.options.options = [
                "None",
                "Model",
                "Value",
            ]
            self.inversion_parameters.reference_model.options.value = "Value"
            flag = self.system.value

            self.inversion_parameters.lower_bound.value = ""
            self.inversion_parameters.upper_bound.value = ""
            self.inversion_parameters.ignore_values.value = "-99999"

        else:
            tx_offsets = self.em_system_specs[self.system.value]["tx_offsets"]
            self.sensor.offset.value = ", ".join(
                [
                    str(offset)
                    for offset in self.em_system_specs[self.system.value]["bird_offset"]
                ]
            )
            uncertainties = self.em_system_specs[self.system.value]["uncertainty"]

            # if start_channel is None:
            start_channel = self.em_system_specs[self.system.value][
                "channel_start_index"
            ]
            if self.em_system_specs[self.system.value]["type"] == "time":
                labels = ["Time (s)"] * len(
                    self.em_system_specs[self.system.value]["channels"]
                )
            else:
                labels = ["Frequency (Hz)"] * len(
                    self.em_system_specs[self.system.value]["channels"]
                )

            system_specs = {}
            for key, time in self.em_system_specs[self.system.value][
                "channels"
            ].items():
                system_specs[key] = f"{time:.5e}"

            self.spatial_choices.options = list(self.spatial_options.keys())

            self.inversion_parameters.reference_model.options.options = [
                "Best-fitting halfspace",
                "Model",
                "Value",
            ]
            self.inversion_parameters.reference_model.options.value = (
                "Best-fitting halfspace"
            )
            self.inversion_parameters.lower_bound.value = "1e-5"
            self.inversion_parameters.upper_bound.value = "10"
            self.inversion_parameters.ignore_values.value = "<0"
            # Switch mesh options
            self.inversion_parameters._mesh = self.mesh_1D
            self.inversion_parameters.inversion_options["mesh"] = self.mesh_1D.widget
            flag = "EM1D"

        self.inversion_parameters.reference_model.value.description = inversion_defaults()[
            "units"
        ][
            flag
        ]
        self.inversion_parameters.reference_model.value.value = inversion_defaults()[
            "reference_value"
        ][flag]
        self.inversion_parameters.reference_model.description.value = (
            "Reference " + inversion_defaults()["property"][flag]
        )
        self.inversion_parameters.starting_model.value.description = inversion_defaults()[
            "units"
        ][
            flag
        ]
        self.inversion_parameters.starting_model.value.value = inversion_defaults()[
            "starting_value"
        ][flag]
        self.inversion_parameters.starting_model.description.value = (
            "Starting " + inversion_defaults()["property"][flag]
        )

        self.spatial_choices.value = self.spatial_choices.options[0]

        def channel_setter(caller):

            channel = caller["owner"]
            data_widget = self.data_channel_choices.data_channel_options[channel.header]

            entity = self.workspace.get_entity(self.selection.objects.value)[0]
            if channel.value is None or not entity.get_data(channel.value):
                data_widget.children[0].value = False
                if self.system.value in ["Magnetics", "Gravity"]:
                    data_widget.children[3].value = "0, 1"
            else:
                data_widget.children[0].value = True
                if self.system.value in ["Magnetics", "Gravity"]:
                    values = entity.get_data(channel.value)[0].values
                    if values is not None and isinstance(values[0], float):
                        data_widget.children[
                            3
                        ].value = f"0, {np.percentile(values[values > 2e-18], 5):.2f}"

            # Trigger plot update
            if self.data_channel_choices.value == channel.header:
                self.plot_selection.plotting_data = channel
                self.plot_selection.refresh.value = False
                self.plot_selection.refresh.value = True

        data_channel_options = {}
        for ind, (key, channel) in enumerate(system_specs.items()):
            if ind + 1 < start_channel:
                continue

            if len(tx_offsets) > 1:
                offsets = tx_offsets[ind]
            else:
                offsets = tx_offsets[0]

            channel_options = ChannelOptions(
                key,
                labels[ind],
                uncertainties=", ".join(
                    [str(uncert) for uncert in uncertainties[ind][:2]]
                ),
                offsets=", ".join([str(offset) for offset in offsets]),
            )

            channel_options.channel_selection.observe(channel_setter, names="value")

            data_channel_options[key] = channel_options.widget

            if self.system.value not in ["Magnetics", "Gravity"]:
                data_channel_options[key].children[1].value = channel
            else:
                data_channel_options[key].children[1].layout.visibility = "hidden"
                data_channel_options[key].children[4].layout.visibility = "hidden"

        if len(data_channel_options) > 0:
            self.data_channel_choices.options = list(data_channel_options.keys())
            self.data_channel_choices.value = list(data_channel_options.keys())[0]
            self.data_channel_choices.data_channel_options = data_channel_options
            self.data_channel_panel.children = [
                self.data_channel_choices,
                data_channel_options[self.data_channel_choices.value],
            ]

        self.update_component_panel()

        if (
            self.system.value not in ["Magnetics", "Gravity"]
            and self.em_system_specs[self.system.value]["type"] == "frequency"
        ):
            self.inversion_parameters.option_choices.options = list(
                self.inversion_parameters.inversion_options.keys()
            )
        else:
            self.inversion_parameters.option_choices.options = [
                key
                for key in self.inversion_parameters.inversion_options.keys()
                if key != "background susceptibility"
            ]

        self.write.button_style = "warning"
        self.run.button_style = "danger"

        if self.system.value == "Magnetics":
            self.survey_type_panel.children = [self.system, self.inducing_field]
        else:
            self.survey_type_panel.children = [self.system]

    def object_observer(self):

        self.plot_selection.resolution.indices = None

        if self.workspace.get_entity(self.selection.objects.value):
            obj = self.workspace.get_entity(self.selection.objects.value)[0]
            data_list = obj.get_data_list()
            # lines.update_list()
            # topography.update_list()

            for aem_system, specs in self.em_system_specs.items():
                if any([specs["flag"] in channel for channel in data_list]):
                    self.system.value = aem_system

            self.system_observer()

            if hasattr(self.data_channel_choices, "data_channel_options"):
                for (
                    key,
                    data_widget,
                ) in self.data_channel_choices.data_channel_options.items():
                    data_widget.children[2].options = self.selection.data.options
                    value = find_value(self.selection.data.options, [key])
                    data_widget.children[2].value = value

            self.write.button_style = "warning"
            self.run.button_style = "danger"

    def get_data_list(self, entity):
        groups = [p_g.name for p_g in entity.property_groups]
        data_list = []
        if self.selection.data.value is not None:
            for component in self.selection.data.value:
                if component in groups:
                    data_list += [
                        self.workspace.get_entity(data)[0].name
                        for data in entity.get_property_group(component).properties
                    ]
                elif component in entity.get_data_list():
                    data_list += [component]
        return data_list

    def update_component_panel(self):
        if self.workspace.get_entity(self.selection.objects.value):
            entity = self.workspace.get_entity(self.selection.objects.value)[0]
            data_list = self.get_data_list(entity)

            if hasattr(self.data_channel_choices, "data_channel_options"):
                for (
                    key,
                    data_widget,
                ) in self.data_channel_choices.data_channel_options.items():
                    data_widget.children[2].options = data_list
                    value = find_value(data_list, [key])
                    data_widget.children[2].value = value

    def data_channel_choices_observer(self):
        if hasattr(
            self.data_channel_choices, "data_channel_options"
        ) and self.data_channel_choices.value in (
            self.data_channel_choices.data_channel_options.keys()
        ):
            data_widget = self.data_channel_choices.data_channel_options[
                self.data_channel_choices.value
            ]
            self.data_channel_panel.children = [self.data_channel_choices, data_widget]

            if (
                self.workspace.get_entity(self.selection.objects.value)
                and data_widget.children[2].value is None
            ):
                entity = self.workspace.get_entity(self.selection.objects.value)[0]
                data_list = self.get_data_list(entity)
                value = find_value(data_list, [self.data_channel_choices.value])
                data_widget.children[2].value = value

        self.write.button_style = "warning"
        self.run.button_style = "danger"

    def spatial_option_change(self):
        self.spatial_panel.children = [
            self.spatial_choices,
            self.spatial_options[self.spatial_choices.value],
        ]
        self.write.button_style = "warning"
        self.run.button_style = "danger"

    def update_octree_param(self):
        dl = self.plot_selection.resolution.value
        self.mesh_octree.core_cell_size.value = f"{dl/2:.0f}, {dl/2:.0f}, {dl/2:.0f}"
        self.mesh_octree.depth_core.value = np.ceil(
            np.min([self.plot_selection.width.value, self.plot_selection.height.value])
            / 2.0
        )
        self.mesh_octree.padding_distance.value = ", ".join(
            list(
                map(
                    str,
                    [
                        np.ceil(self.plot_selection.width.value / 2),
                        np.ceil(self.plot_selection.width.value / 2),
                        np.ceil(self.plot_selection.height.value / 2),
                        np.ceil(self.plot_selection.height.value / 2),
                        0,
                        0,
                    ],
                )
            )
        )
        self.plot_selection.resolution.indices = None
        self.write.button_style = "warning"
        self.run.button_style = "danger"

    def update_selection(self):
        self.plot_selection.highlight_selection = {
            self.lines.data.value: self.lines.lines.value
        }
        self.plot_selection.refresh.value = False
        self.plot_selection.refresh.value = True

    def write_trigger(self):
        if self.write.value is False:
            return

        input_dict = {
            "out_group": self.inversion_parameters.output_name.value,
            "workspace": self.h5file,
            "save_to_geoh5": self.h5file,
        }
        if self.system.value in ["Gravity", "Magnetics"]:
            input_dict["inversion_type"] = self.system.value.lower()

            if input_dict["inversion_type"] == "magnetics":
                input_dict["inducing_field_aid"] = string_2_list(
                    self.inducing_field.value
                )
            # Octree mesh parameters
            input_dict["core_cell_size"] = string_2_list(
                self.mesh_octree.core_cell_size.value
            )
            input_dict["octree_levels_topo"] = string_2_list(
                self.mesh_octree.octree_levels_topo.value
            )
            input_dict["octree_levels_obs"] = string_2_list(
                self.mesh_octree.octree_levels_obs.value
            )
            input_dict["depth_core"] = {"value": self.mesh_octree.depth_core.value}
            input_dict["max_distance"] = self.mesh_octree.max_distance.value
            p_d = string_2_list(self.mesh_octree.padding_distance.value)
            input_dict["padding_distance"] = [
                [p_d[0], p_d[1]],
                [p_d[2], p_d[3]],
                [p_d[4], p_d[5]],
            ]

        else:
            input_dict["system"] = self.system.value
            input_dict["lines"] = {
                self.lines.data.value: [str(line) for line in self.lines.lines.value]
            }

            input_dict["mesh 1D"] = [
                self.mesh_1D.hz_min.value,
                self.mesh_1D.hz_expansion.value,
                self.mesh_1D.n_cells.value,
            ]
        input_dict["chi_factor"] = self.inversion_parameters.chi_factor.value
        input_dict["max_iterations"] = self.inversion_parameters.max_iterations.value
        input_dict[
            "max_cg_iterations"
        ] = self.inversion_parameters.max_cg_iterations.value

        if self.inversion_parameters.beta_start_options.value == "value":
            input_dict["initial_beta"] = self.inversion_parameters.beta_start.value
        else:
            input_dict[
                "initial_beta_ratio"
            ] = self.inversion_parameters.beta_start.value

        input_dict["tol_cg"] = self.inversion_parameters.tol_cg.value
        input_dict["ignore_values"] = self.inversion_parameters.ignore_values.value
        input_dict["resolution"] = self.plot_selection.resolution.value
        input_dict["window"] = {
            "center": [
                self.plot_selection.center_x.value,
                self.plot_selection.center_y.value,
            ],
            "size": [self.plot_selection.width.value, self.plot_selection.height.value],
            "azimuth": self.plot_selection.azimuth.value,
        }
        input_dict["alphas"] = string_2_list(self.inversion_parameters.alphas.value)
        input_dict["reference_model"] = {
            self.inversion_parameters.reference_model.widget.children[1]
            .children[0]
            .value.lower(): self.inversion_parameters.reference_model.widget.children[1]
            .children[1]
            .value
        }
        input_dict["starting_model"] = {
            self.inversion_parameters.starting_model.widget.children[1]
            .children[0]
            .value.lower(): self.inversion_parameters.starting_model.widget.children[1]
            .children[1]
            .value
        }

        if self.inversion_parameters.susceptibility_model.options.value != "None":
            input_dict["susceptibility"] = {
                self.inversion_parameters.susceptibility_model.widget.children[1]
                .children[0]
                .value.lower(): self.inversion_parameters.susceptibility_model.widget.children[
                    1
                ]
                .children[1]
                .value
            }

        input_dict["model_norms"] = string_2_list(self.inversion_parameters.norms.value)

        if len(self.inversion_parameters.lower_bound.value) > 1:
            input_dict["lower_bound"] = string_2_list(
                self.inversion_parameters.lower_bound.value
            )

        if len(self.inversion_parameters.upper_bound.value) > 1:
            input_dict["upper_bound"] = string_2_list(
                self.inversion_parameters.upper_bound.value
            )

        input_dict["data"] = {}
        input_dict["data"]["type"] = "GA_object"
        input_dict["data"]["name"] = self.selection.objects.value

        if hasattr(self.data_channel_choices, "data_channel_options"):
            channel_param = {}

            for (
                key,
                data_widget,
            ) in self.data_channel_choices.data_channel_options.items():
                if data_widget.children[0].value is False:
                    continue

                channel_param[key] = {}
                channel_param[key]["name"] = data_widget.children[2].value
                channel_param[key]["uncertainties"] = string_2_list(
                    data_widget.children[3].value
                )
                channel_param[key]["offsets"] = string_2_list(
                    data_widget.children[4].value
                )

                if self.system.value not in ["Gravity", "Magnetics"]:
                    channel_param[key]["value"] = string_2_list(
                        data_widget.children[1].value
                    )
                if (
                    self.system.value in ["Gravity", "Magnetics"]
                    and self.plot_selection.azimuth.value != 0
                    and key not in ["tmi", "gz"]
                ):
                    print(
                        f"Gradient data with rotated window is currently not supported"
                    )
                    self.run.button_style = "danger"
                    return

            input_dict["data"]["channels"] = channel_param

        input_dict["uncertainty_mode"] = self.inversion_parameters.uncert_mode.value

        if self.sensor.options_button.value == "(x, y, z) + offset(x,y,z)":
            input_dict["receivers_offset"] = {
                "constant": string_2_list(self.sensor.offset.value)
            }
        else:
            input_dict["receivers_offset"] = {
                "radar_drape": string_2_list(self.sensor.offset.value)
                + [self.sensor.value.value]
            }

        if self.topography.options_button.value == "Object":
            if self.topography.objects.value is None:
                input_dict["topography"] = None
            else:
                input_dict["topography"] = {
                    "GA_object": {
                        "name": self.topography.objects.value,
                        "data": self.topography.value.value,
                    }
                }
        elif self.topography.options_button.value == "Drape Height":
            input_dict["topography"] = {"drapped": self.topography.offset.value}
        else:
            input_dict["topography"] = {"constant": self.topography.constant.value}

        if self.forward_only.value:
            input_dict["forward_only"] = []

        checks = [key for key, val in input_dict.items() if val is None]

        if len(list(input_dict["data"]["channels"].keys())) == 0:
            checks += ["'Channel' for at least one data component."]

        if len(checks) > 0:
            print(f"Required value for {checks}")
            self.run.button_style = "danger"
        else:
            self.write.button_style = ""
            file = self._inv_dir + f"{self.inversion_parameters.output_name.value}.json"
            with open(file, "w") as f:
                json.dump(input_dict, f, indent=4)
            self.run.button_style = "success"

        self.write.value = False
        self.write.button_style = ""
        self.run.button_style = "success"

    # for attr, item in kwargs.items():
    #     try:
    #         if hasattr(w_l[attr], "options") and item not in w_l[attr].options:
    #             continue
    #         w_l[attr].value = item
    #     except KeyError:
    #         continue
    # for attr, item in kwargs.items():
    #     try:
    #         if getattr(self.selection, attr, None) is not None:
    #             widget = getattr(self.selection, attr)
    #             if hasattr(widget, "options") and item not in widget.options:
    #                 continue
    #             widget.value = item
    #     except AttributeError:
    #         continue



""" module SimPEG.EM.NSEM

SimPEG implementation of the natural source problem
(including magenetotelluric, tipper and ZTEM)



"""

from . import Utils
from . import SrcNSEM as Src
from . import RxNSEM as Rx
from .SurveyNSEM import Survey, Data
from .FieldsNSEM import Fields1D_ePrimSec, Fields3D_ePrimSec
from .ProblemNSEM import Problem1D_ePrimSec, Problem3D_ePrimSec



# ------------ Dummy codes for older version ------------ #
# This can be used in later use for handling on-time data.
# ------------------------------------------------------- #

# def setWaveform(self, **kwargs):
#     """
#         Set parameters for Src Waveform
#     """
#     # TODO: this is hp is only valid for Circular loop system
#     self.hp = self.I/self.a*0.5

#     self.toff = kwargs['toff']
#     self.waveform = kwargs['waveform']
#     self.waveformDeriv = kwargs['waveformDeriv']
#     self.tconv = kwargs['tconv']

# def projectFields(self, u):
#     """
#         Transform frequency domain responses to time domain responses
#     """
#     # Case1: Compute frequency domain reponses right at filter coefficient values
#     if self.switchInterp == False:
#         # Src waveform: Step-off
#         if self.wave_type == 'stepoff':
#             if self.rx_type == 'Bz':
#                 # Compute EM responses
#                 if u.size == self.n_frequency:
#                     resp, f0 = transFilt(Utils.mkvc(u), self.wt,self.tbase, self.frequency*2*np.pi, self.time)
#                 # Compute EM sensitivities
#                 else:
#                     resp = np.zeros((self.n_time, self.n_layer))
#                     for i in range (self.n_layer):
#                         resp[:,i], f0 = transFilt(u[:,i], self.wt,self.tbase, self.frequency*2*np.pi, self.time)

#             elif self.rx_type == 'dBzdt':
#                 # Compute EM responses
#                 if u.size == self.n_frequency:
#                     resp = -transFiltImpulse(u, self.wt,self.tbase, self.frequency*2*np.pi, self.time)
#                 # Compute EM sensitivities
#                 else:
#                     resp = np.zeros((self.n_time, self.n_layer))
#                     for i in range (self.n_layer):
#                         resp[:,i] = -transFiltImpulse(u[:,i], self.wt,self.tbase, self.frequency*2*np.pi, self.time)

#         # Src waveform: General (it can be any waveform)
#         # We evaluate this with time convolution
#         elif self.wave_type == 'general':
#             # Compute EM responses
#             if u.size == self.n_frequency:
#                 # TODO: write small code which compute f at t = 0
#                 f, f0 = transFilt(Utils.mkvc(u), self.wt, self.tbase, self.frequency*2*np.pi, self.tconv)
#                 fDeriv = -transFiltImpulse(Utils.mkvc(u), self.wt,self.tbase, self.frequency*2*np.pi, self.tconv)

#                 if self.rx_type == 'Bz':

#                     waveConvfDeriv = CausalConv(self.waveform, fDeriv, self.tconv)
#                     resp1 = (self.waveform*self.hp*(1-f0[1]/self.hp)) - waveConvfDeriv
#                     respint = interp1d(self.tconv, resp1, 'linear')

#                     # TODO: make it as an opition #2
#                     # waveDerivConvf = CausalConv(self.waveformDeriv, f, self.tconv)
#                     # resp2 = (self.waveform*self.hp) - waveDerivConvf
#                     # respint = interp1d(self.tconv, resp2, 'linear')

#                     resp = respint(self.time)

#                 if self.rx_type == 'dBzdt':
#                     waveDerivConvfDeriv = CausalConv(self.waveformDeriv, fDeriv, self.tconv)
#                     resp1 = self.hp*self.waveformDeriv*(1-f0[1]/self.hp) - waveDerivConvfDeriv
#                     respint = interp1d(self.tconv, resp1, 'linear')
#                     resp = respint(self.time)

#             # Compute EM sensitivities
#             else:

#                 resp = np.zeros((self.n_time, self.n_layer))
#                 for i in range (self.n_layer):

#                     f, f0 = transFilt(u[:,i], self.wt, self.tbase, self.frequency*2*np.pi, self.tconv)
#                     fDeriv = -transFiltImpulse(u[:,i], self.wt,self.tbase, self.frequency*2*np.pi, self.tconv)

#                     if self.rx_type == 'Bz':

#                         waveConvfDeriv = CausalConv(self.waveform, fDeriv, self.tconv)
#                         resp1 = (self.waveform*self.hp*(1-f0[1]/self.hp)) - waveConvfDeriv
#                         respint = interp1d(self.tconv, resp1, 'linear')

#                         # TODO: make it as an opition #2
#                         # waveDerivConvf = CausalConv(self.waveformDeriv, f, self.tconv)
#                         # resp2 = (self.waveform*self.hp) - waveDerivConvf
#                         # respint = interp1d(self.tconv, resp2, 'linear')

#                         resp[:,i] = respint(self.time)

#                     if self.rx_type == 'dBzdt':
#                         waveDerivConvfDeriv = CausalConv(self.waveformDeriv, fDeriv, self.tconv)
#                         resp1 = self.hp*self.waveformDeriv*(1-f0[1]/self.hp) - waveDerivConvfDeriv
#                         respint = interp1d(self.tconv, resp1, 'linear')
#                         resp[:,i] = respint(self.time)

#     # Case2: Compute frequency domain reponses in logarithmic then intepolate
#     if self.switchInterp == True:
#         # Src waveform: Step-off
#         if self.wave_type == 'stepoff':
#             if self.rx_type == 'Bz':
#                 # Compute EM responses
#                 if u.size == self.n_frequency:
#                     resp, f0 = transFiltInterp(Utils.mkvc(u), self.wt,self.tbase, self.frequency*2*np.pi, self.omega_int, self.time)
#                 # Compute EM sensitivities
#                 else:
#                     resp = np.zeros((self.n_time, self.n_layer))
#                     for i in range (self.n_layer):
#                         resp[:,i], f0 = transFiltInterp(u[:,i], self.wt,self.tbase, self.frequency*2*np.pi, self.omega_int, self.time)

#             elif self.rx_type == 'dBzdt':
#                 # Compute EM responses
#                 if u.size == self.n_frequency:
#                     resp = -transFiltImpulseInterp(Utils.mkvc(u), self.wt,self.tbase, self.frequency*2*np.pi, self.omega_int, self.time)
#                 # Compute EM sensitivities
#                 else:
#                     resp = np.zeros((self.n_time, self.n_layer))
#                     for i in range (self.n_layer):
#                         resp[:,i] = -transFiltImpulseInterp(u[:,i], self.wt,self.tbase, self.frequency*2*np.pi, self.omega_int, self.time)

#         # Src waveform: General (it can be any waveform)
#         # We evaluate this with time convolution
#         elif self.wave_type == 'general':
#             # Compute EM responses
#             if u.size == self.n_frequency:
#                 # TODO: write small code which compute f at t = 0
#                 f, f0 = transFiltInterp(Utils.mkvc(u), self.wt, self.tbase, self.frequency*2*np.pi, self.omega_int, self.tconv)
#                 fDeriv = -transFiltImpulseInterp(Utils.mkvc(u), self.wt,self.tbase, self.frequency*2*np.pi, self.omega_int, self.tconv)

#                 if self.rx_type == 'Bz':

#                     waveConvfDeriv = CausalConv(self.waveform, fDeriv, self.tconv)
#                     resp1 = (self.waveform*self.hp*(1-f0[1]/self.hp)) - waveConvfDeriv
#                     respint = interp1d(self.tconv, resp1, 'linear')

#                     # TODO: make it as an opition #2
#                     # waveDerivConvf = CausalConv(self.waveformDeriv, f, self.tconv)
#                     # resp2 = (self.waveform*self.hp) - waveDerivConvf
#                     # respint = interp1d(self.tconv, resp2, 'linear')

#                     resp = respint(self.time)

#                 if self.rx_type == 'dBzdt':
#                     waveDerivConvfDeriv = CausalConv(self.waveformDeriv, fDeriv, self.tconv)
#                     resp1 = self.hp*self.waveformDeriv*(1-f0[1]/self.hp) - waveDerivConvfDeriv
#                     respint = interp1d(self.tconv, resp1, 'linear')
#                     resp = respint(self.time)

#             # Compute EM sensitivities
#             else:

#                 resp = np.zeros((self.n_time, self.n_layer))
#                 for i in range (self.n_layer):

#                     f, f0 = transFiltInterp(u[:,i], self.wt, self.tbase, self.frequency*2*np.pi, self.omega_int, self.tconv)
#                     fDeriv = -transFiltImpulseInterp(u[:,i], self.wt,self.tbase, self.frequency*2*np.pi, self.omega_int, self.tconv)

#                     if self.rx_type == 'Bz':

#                         waveConvfDeriv = CausalConv(self.waveform, fDeriv, self.tconv)
#                         resp1 = (self.waveform*self.hp*(1-f0[1]/self.hp)) - waveConvfDeriv
#                         respint = interp1d(self.tconv, resp1, 'linear')

#                         # TODO: make it as an opition #2
#                         # waveDerivConvf = CausalConv(self.waveformDeriv, f, self.tconv)
#                         # resp2 = (self.waveform*self.hp) - waveDerivConvf
#                         # respint = interp1d(self.tconv, resp2, 'linear')

#                         resp[:,i] = respint(self.time)

#                     if self.rx_type == 'dBzdt':
#                         waveDerivConvfDeriv = CausalConv(self.waveformDeriv, fDeriv, self.tconv)
#                         resp1 = self.hp*self.waveformDeriv*(1-f0[1]/self.hp) - waveDerivConvfDeriv
#                         respint = interp1d(self.tconv, resp1, 'linear')
#                         resp[:,i] = respint(self.time)

#     return mu_0*resp



from .. import Utils
from SimPEG.EM.Base import BaseEMProblem
from .SurveyDC import Survey
from .FieldsDC import FieldsDC, Fields_CC, Fields_N
import numpy as np
import scipy as sp
from SimPEG.Utils import Zero
from .BoundaryUtils import getxBCyBC_CC


class BaseDCProblem(BaseEMProblem):
    """
    Base DC Problem
    """

    surveyPair = Survey
    fieldsPair = FieldsDC
    Ainv = None

    def fields(self, m=None):
        if m is not None:
            self.model = m

        if self.Ainv is not None:
            self.Ainv.clean()

        f = self.fieldsPair(self.mesh, self.survey)
        A = self.getA()
        self.Ainv = self.Solver(A, **self.solverOpts)
        RHS = self.getRHS()
        u = self.Ainv * RHS
        Srcs = self.survey.srcList
        f[Srcs, self._solutionType] = u
        return f

    def Jvec(self, m, v, f=None):

        if f is None:
            f = self.fields(m)

        self.model = m

        # Jv = self.dataPair(self.survey)  # same size as the data
        Jv = []

        A = self.getA()

        for src in self.survey.srcList:
            u_src = f[src, self._solutionType]  # solution vector
            dA_dm_v = self.getADeriv(u_src, v)
            dRHS_dm_v = self.getRHSDeriv(src, v)
            du_dm_v = self.Ainv * (-dA_dm_v + dRHS_dm_v)

            for rx in src.rxList:
                df_dmFun = getattr(f, f"_{rx.projField!s}Deriv", None)
                df_dm_v = df_dmFun(src, du_dm_v, v, adjoint=False)
                Jv.append(rx.evalDeriv(src, self.mesh, f, df_dm_v))
                # Jv[src, rx] = rx.evalDeriv(src, self.mesh, f, df_dm_v)
        # return Utils.mkvc(Jv)
        return np.hstack(Jv)

    def Jtvec(self, m, v, f=None):
        if f is None:
            f = self.fields(m)

        self.model = m

        # Ensure v is a data object.
        if not isinstance(v, self.dataPair):
            v = self.dataPair(self.survey, v)

        Jtv = np.zeros(m.size)
        AT = self.getA()

        for src in self.survey.srcList:
            u_src = f[src, self._solutionType]
            for rx in src.rxList:
                # wrt f, need possibility wrt m
                PTv = rx.evalDeriv(src, self.mesh, f, v[src, rx], adjoint=True)
                df_duTFun = getattr(f, f"_{rx.projField!s}Deriv", None)
                df_duT, df_dmT = df_duTFun(src, None, PTv, adjoint=True)

                ATinvdf_duT = self.Ainv * df_duT

                dA_dmT = self.getADeriv(u_src, ATinvdf_duT, adjoint=True)
                dRHS_dmT = self.getRHSDeriv(src, ATinvdf_duT, adjoint=True)
                du_dmT = -dA_dmT + dRHS_dmT
                Jtv += (df_dmT + du_dmT).astype(float)

        return Utils.mkvc(Jtv)

    def getSourceTerm(self):
        """
        Evaluates the sources, and puts them in matrix form

        :rtype: tuple
        :return: q (nC or nN, nSrc)
        """

        Srcs = self.survey.srcList

        if self._formulation == "EB":
            n = self.mesh.nN
            # return NotImplementedError

        elif self._formulation == "HJ":
            n = self.mesh.nC

        q = np.zeros((n, len(Srcs)))

        for i, src in enumerate(Srcs):
            q[:, i] = src.eval(self)
        return q


class Problem3D_CC(BaseDCProblem):
    """
    3D cell centered DC problem
    """

    _solutionType = "phiSolution"
    _formulation = "HJ"  # CC potentials means J is on faces
    fieldsPair = Fields_CC
    bc_type = "Neumann"

    def __init__(self, mesh, **kwargs):

        BaseDCProblem.__init__(self, mesh, **kwargs)
        self.setBC()

    def getA(self):
        """

        Make the A matrix for the cell centered DC resistivity problem

        A = D MfRhoI G

        """

        D = self.Div
        G = self.Grad
        MfRhoI = self.MfRhoI
        A = D * MfRhoI * G

        if self.bc_type == "Neumann":
            Vol = self.mesh.vol
            if self.verbose:
                print("Perturbing first row of A to remove nullspace for Neumann BC.")

            # Handling Null space of A
            I, J, V = sp.sparse.find(A[0, :])
            for jj in J:
                A[0, jj] = 0.0

            A[0, 0] = 1.0 / Vol[0]

        # I think we should deprecate this for DC problem.
        # if self._makeASymmetric is True:
        #     return V.T * A
        return A

    def getADeriv(self, u, v, adjoint=False):

        D = self.Div
        G = self.Grad
        MfRhoIDeriv = self.MfRhoIDeriv

        if adjoint:
            return (MfRhoIDeriv(G * u).T) * (D.T * v)

        return D * (MfRhoIDeriv(G * u) * v)

    def getRHS(self):
        """
        RHS for the DC problem

        q
        """

        RHS = self.getSourceTerm()

        return RHS

    def getRHSDeriv(self, src, v, adjoint=False):
        """
        Derivative of the right hand side with respect to the model
        """
        # TODO: add qDeriv for RHS depending on m
        # qDeriv = src.evalDeriv(self, adjoint=adjoint)
        # return qDeriv
        return Zero()

    def setBC(self):
        if self.mesh._meshType == "TREE":
            if self.bc_type == "Neumann":
                raise NotImplementedError()
            elif self.bc_type == "Dirchlet":
                print(
                    "Homogeneous Dirchlet is the natural BC for this CC discretization."
                )
                self.Div = Utils.sdiag(self.mesh.vol) * self.mesh.faceDiv
                self.Grad = self.Div.T

        else:

            if self.mesh.dim == 3:
                fxm, fxp, fym, fyp, fzm, fzp = self.mesh.faceBoundaryInd
                gBFxm = self.mesh.gridFx[fxm, :]
                gBFxp = self.mesh.gridFx[fxp, :]
                gBFym = self.mesh.gridFy[fym, :]
                gBFyp = self.mesh.gridFy[fyp, :]
                gBFzm = self.mesh.gridFz[fzm, :]
                gBFzp = self.mesh.gridFz[fzp, :]

                # Setup Mixed B.C (alpha, beta, gamma)
                temp_xm = np.ones_like(gBFxm[:, 0])
                temp_xp = np.ones_like(gBFxp[:, 0])
                temp_ym = np.ones_like(gBFym[:, 1])
                temp_yp = np.ones_like(gBFyp[:, 1])
                temp_zm = np.ones_like(gBFzm[:, 2])
                temp_zp = np.ones_like(gBFzp[:, 2])

                if self.bc_type == "Neumann":
                    if self.verbose:
                        print("Setting BC to Neumann.")
                    alpha_xm, alpha_xp = temp_xm * 0.0, temp_xp * 0.0
                    alpha_ym, alpha_yp = temp_ym * 0.0, temp_yp * 0.0
                    alpha_zm, alpha_zp = temp_zm * 0.0, temp_zp * 0.0

                    beta_xm, beta_xp = temp_xm, temp_xp
                    beta_ym, beta_yp = temp_ym, temp_yp
                    beta_zm, beta_zp = temp_zm, temp_zp

                    gamma_xm, gamma_xp = temp_xm * 0.0, temp_xp * 0.0
                    gamma_ym, gamma_yp = temp_ym * 0.0, temp_yp * 0.0
                    gamma_zm, gamma_zp = temp_zm * 0.0, temp_zp * 0.0

                elif self.bc_type == "Dirchlet":
                    if self.verbose:
                        print("Setting BC to Dirchlet.")
                    alpha_xm, alpha_xp = temp_xm, temp_xp
                    alpha_ym, alpha_yp = temp_ym, temp_yp
                    alpha_zm, alpha_zp = temp_zm, temp_zp

                    beta_xm, beta_xp = temp_xm * 0, temp_xp * 0
                    beta_ym, beta_yp = temp_ym * 0, temp_yp * 0
                    beta_zm, beta_zp = temp_zm * 0, temp_zp * 0

                    gamma_xm, gamma_xp = temp_xm * 0.0, temp_xp * 0.0
                    gamma_ym, gamma_yp = temp_ym * 0.0, temp_yp * 0.0
                    gamma_zm, gamma_zp = temp_zm * 0.0, temp_zp * 0.0

                alpha = [alpha_xm, alpha_xp, alpha_ym, alpha_yp, alpha_zm, alpha_zp]
                beta = [beta_xm, beta_xp, beta_ym, beta_yp, beta_zm, beta_zp]
                gamma = [gamma_xm, gamma_xp, gamma_ym, gamma_yp, gamma_zm, gamma_zp]

            elif self.mesh.dim == 2:

                fxm, fxp, fym, fyp = self.mesh.faceBoundaryInd
                gBFxm = self.mesh.gridFx[fxm, :]
                gBFxp = self.mesh.gridFx[fxp, :]
                gBFym = self.mesh.gridFy[fym, :]
                gBFyp = self.mesh.gridFy[fyp, :]

                # Setup Mixed B.C (alpha, beta, gamma)
                temp_xm = np.ones_like(gBFxm[:, 0])
                temp_xp = np.ones_like(gBFxp[:, 0])
                temp_ym = np.ones_like(gBFym[:, 1])
                temp_yp = np.ones_like(gBFyp[:, 1])

                alpha_xm, alpha_xp = temp_xm * 0.0, temp_xp * 0.0
                alpha_ym, alpha_yp = temp_ym * 0.0, temp_yp * 0.0

                beta_xm, beta_xp = temp_xm, temp_xp
                beta_ym, beta_yp = temp_ym, temp_yp

                gamma_xm, gamma_xp = temp_xm * 0.0, temp_xp * 0.0
                gamma_ym, gamma_yp = temp_ym * 0.0, temp_yp * 0.0

                alpha = [alpha_xm, alpha_xp, alpha_ym, alpha_yp]
                beta = [beta_xm, beta_xp, beta_ym, beta_yp]
                gamma = [gamma_xm, gamma_xp, gamma_ym, gamma_yp]

            x_BC, y_BC = getxBCyBC_CC(self.mesh, alpha, beta, gamma)
            V = self.Vol
            self.Div = V * self.mesh.faceDiv
            P_BC, B = self.mesh.getBCProjWF_simple()
            M = B * self.mesh.aveCC2F
            self.Grad = self.Div.T - P_BC * Utils.sdiag(y_BC) * M


class Problem3D_N(BaseDCProblem):
    """
    3D nodal DC problem
    """

    _solutionType = "phiSolution"
    _formulation = "EB"  # N potentials means B is on faces
    fieldsPair = Fields_N

    def __init__(self, mesh, **kwargs):
        BaseDCProblem.__init__(self, mesh, **kwargs)

    def getA(self):
        """

        Make the A matrix for the cell centered DC resistivity problem

        A = G.T MeSigma G

        """

        MeSigma = self.MeSigma
        Grad = self.mesh.nodalGrad
        A = Grad.T * MeSigma * Grad

        Vol = self.mesh.vol

        # Handling Null space of A
        I, J, V = sp.sparse.find(A[0, :])
        for jj in J:
            A[0, jj] = 0.0

        A[0, 0] = 1.0 / Vol[0]

        return A

    def getADeriv(self, u, v, adjoint=False):
        """

        Product of the derivative of our system matrix with respect to the
        model and a vector

        """
        Grad = self.mesh.nodalGrad
        if not adjoint:
            return Grad.T * (self.MeSigmaDeriv(Grad * u) * v)
        elif adjoint:
            return self.MeSigmaDeriv(Grad * u).T * (Grad * v)

    def getRHS(self):
        """
        RHS for the DC problem

        q
        """

        RHS = self.getSourceTerm()
        return RHS

    def getRHSDeriv(self, src, v, adjoint=False):
        """
        Derivative of the right hand side with respect to the model
        """
        # TODO: add qDeriv for RHS depending on m
        # qDeriv = src.evalDeriv(self, adjoint=adjoint)
        # return qDeriv
        return Zero()



from matplotlib import pyplot as plt, colors, numpy as np


def plotIsoFreqNSimpedance(
    ax,
    freq,
    array,
    flag,
    par="abs",
    colorbar=True,
    colorNorm="SymLog",
    cLevel=True,
    contour=True,
):

    indUniFreq = np.where(freq == array["freq"])

    x, y = array["x"][indUniFreq], array["y"][indUniFreq]
    if par == "abs":
        zPlot = np.abs(array[flag][indUniFreq])
        cmap = plt.get_cmap("OrRd_r")  # seismic')
        level = np.logspace(0, -5, 31)
        clevel = np.logspace(0, -4, 5)
        plotNorm = colors.LogNorm()
    elif par == "real":
        zPlot = np.real(array[flag][indUniFreq])
        cmap = plt.get_cmap("RdYlBu")
        if cLevel:
            level = np.concatenate((-np.logspace(0, -10, 31), np.logspace(-10, 0, 31)))
            clevel = np.concatenate((-np.logspace(0, -8, 5), np.logspace(-8, 0, 5)))
        else:
            level = np.linspace(zPlot.min(), zPlot.max(), 100)
            clevel = np.linspace(zPlot.min(), zPlot.max(), 10)
        if colorNorm == "SymLog":
            plotNorm = colors.SymLogNorm(1e-10, linscale=2)
        else:
            plotNorm = colors.Normalize()
    elif par == "imag":
        zPlot = np.imag(array[flag][indUniFreq])
        cmap = plt.get_cmap("RdYlBu")
        level = np.concatenate((-np.logspace(0, -10, 31), np.logspace(-10, 0, 31)))
        clevel = np.concatenate((-np.logspace(0, -8, 5), np.logspace(-8, 0, 5)))
        plotNorm = colors.SymLogNorm(1e-10, linscale=2)
        if cLevel:
            level = np.concatenate((-np.logspace(0, -10, 31), np.logspace(-10, 0, 31)))
            clevel = np.concatenate((-np.logspace(0, -8, 5), np.logspace(-8, 0, 5)))
        else:
            level = np.linspace(zPlot.min(), zPlot.max(), 100)
            clevel = np.linspace(zPlot.min(), zPlot.max(), 10)
        if colorNorm == "SymLog":
            plotNorm = colors.SymLogNorm(1e-10, linscale=2)
        elif colorNorm == "Lin":
            plotNorm = colors.Normalize()
    if contour:
        cs = ax.tricontourf(
            x, y, zPlot, levels=level, cmap=cmap, norm=plotNorm
        )  # ,extend='both')
    else:
        uniX, uniY = np.unique(x), np.unique(y)
        X, Y = np.meshgrid(
            np.append(uniX - 25, uniX[-1] + 25), np.append(uniY - 25, uniY[-1] + 25)
        )
        cs = ax.pcolor(
            X, Y, np.reshape(zPlot, (len(uniY), len(uniX))), cmap=cmap, norm=plotNorm
        )
    if colorbar:
        plt.colorbar(cs, cax=ax.cax, ticks=clevel, format="%1.2e")
        ax.set_title(flag + " " + par, fontsize=8)
    return cs


def plotIsoFreqNSDiff(
    ax,
    freq,
    arrayList,
    flag,
    par="abs",
    colorbar=True,
    cLevel=True,
    mask=None,
    contourLine=True,
    useLog=False,
):

    indUniFreq0 = np.where(freq == arrayList[0]["freq"])
    indUniFreq1 = np.where(freq == arrayList[1]["freq"])
    seicmap = plt.get_cmap("RdYlBu")  # seismic')
    x, y = arrayList[0]["x"][indUniFreq0], arrayList[0]["y"][indUniFreq0]
    if par == "abs":
        if useLog:
            zPlot = (
                np.log10(np.abs(arrayList[0][flag][indUniFreq0]))
                - np.log10(np.abs(arrayList[1][flag][indUniFreq1]))
            ) / np.log10(np.abs(arrayList[1][flag][indUniFreq1]))
        else:
            zPlot = (
                np.abs(arrayList[0][flag][indUniFreq0])
                - np.abs(arrayList[1][flag][indUniFreq1])
            ) / np.abs(arrayList[1][flag][indUniFreq1])
        if mask:
            maskInd = np.logical_or(
                np.abs(arrayList[0][flag][indUniFreq0]) < 1e-3,
                np.abs(arrayList[1][flag][indUniFreq1]) < 1e-3,
            )
            zPlot = np.ma.array(zPlot)
            zPlot[maskInd] = mask
        if cLevel:
            level = np.arange(-200, 201, 10)
            clevel = np.arange(-200, 201, 25)
        else:
            level = np.linspace(zPlot.min(), zPlot.max(), 100)
            clevel = np.linspace(zPlot.min(), zPlot.max(), 10)
    elif par == "real":
        if useLog:
            zPlot = (
                np.log10(np.real(arrayList[0][flag][indUniFreq0]))
                - np.log10(np.real(arrayList[1][flag][indUniFreq1]))
            ) / np.log10(np.abs(np.real(arrayList[1][flag][indUniFreq1])))
        else:
            zPlot = (
                np.real(arrayList[0][flag][indUniFreq0])
                - np.real(arrayList[1][flag][indUniFreq1])
            ) / np.abs(np.real(arrayList[1][flag][indUniFreq1]))
        if mask:
            maskInd = np.logical_or(
                np.abs(np.real(arrayList[0][flag][indUniFreq0])) < 1e-3,
                np.abs(np.real(arrayList[1][flag][indUniFreq1])) < 1e-3,
            )
            zPlot = np.ma.array(zPlot)
            zPlot[maskInd] = mask
        if cLevel:
            level = np.arange(-200, 201, 10)
            clevel = np.arange(-200, 201, 25)
        else:
            level = np.linspace(zPlot.min(), zPlot.max(), 100)
            clevel = np.linspace(zPlot.min(), zPlot.max(), 10)
    elif par == "imag":
        if useLog:
            zPlot = (
                np.log10(np.imag(arrayList[0][flag][indUniFreq0]))
                - np.log10(np.imag(arrayList[1][flag][indUniFreq1]))
            ) / np.log10(np.abs(np.imag(arrayList[1][flag][indUniFreq1])))
        else:
            zPlot = (
                np.imag(arrayList[0][flag][indUniFreq0])
                - np.imag(arrayList[1][flag][indUniFreq1])
            ) / np.abs(np.imag(arrayList[1][flag][indUniFreq1]))
        if mask:
            maskInd = np.logical_or(
                np.abs(np.imag(arrayList[0][flag][indUniFreq0])) < 1e-3,
                np.abs(np.imag(arrayList[1][flag][indUniFreq1])) < 1e-3,
            )
            zPlot = np.ma.array(zPlot)
            zPlot[maskInd] = mask
        if cLevel:
            level = np.arange(-200, 201, 10)
            clevel = np.arange(-200, 201, 25)
        else:
            level = np.linspace(zPlot.min(), zPlot.max(), 100)
            clevel = np.linspace(zPlot.min(), zPlot.max(), 10)
    cs = ax.tricontourf(
        x, y, zPlot * 100, levels=level * 100, cmap=seicmap, extend="both"
    )  # ,norm=colors.SymLogNorm(1e-2,linscale=2))
    if contourLine:
        csl = ax.tricontour(x, y, zPlot * 100, levels=clevel * 100, colors="k")
        plt.clabel(csl, fontsize=7, inline=1, fmt="%1.1e", inline_spacing=10)
    if colorbar:
        cb = plt.colorbar(cs, cax=ax.cax, ticks=clevel * 100, format="%1.1e")
        for t in cb.ax.get_yticklabels():
            t.set_rotation(60)
            t.set_fontsize(8)

    ax.set_title(flag + " " + par, fontsize=8)


def plotIsoFreqNStipper(
    ax,
    freq,
    array,
    flag,
    par="abs",
    colorbar=True,
    colorNorm="SymLog",
    cLevel=True,
    contour=True,
):

    indUniFreq = np.where(freq == array["freq"])

    x, y = array["x"][indUniFreq], array["y"][indUniFreq]
    if par == "abs":
        cmap = plt.get_cmap("OrRd_r")  # seismic')
        zPlot = np.abs(array[flag][indUniFreq])
        if cLevel:
            level = np.logspace(-4, 0, 33)
            clevel = np.logspace(-4, 0, 5)
        else:
            level = np.linspace(zPlot.min(), zPlot.max(), 100)
            clevel = np.linspace(zPlot.min(), zPlot.max(), 10)
        if colorNorm == "SymLog":
            plotNorm = colors.LogNorm()
        else:
            plotNorm = colors.Normalize()
    elif par == "real":
        cmap = plt.get_cmap("RdYlBu")
        zPlot = np.real(array[flag][indUniFreq])
        if cLevel:
            level = np.concatenate((-np.logspace(0, -4, 33), np.logspace(-4, 0, 33)))
            clevel = np.concatenate((-np.logspace(0, -4, 5), np.logspace(-4, 0, 5)))
        else:
            level = np.linspace(zPlot.min(), zPlot.max(), 100)
            clevel = np.linspace(zPlot.min(), zPlot.max(), 10)
        if colorNorm == "SymLog":
            plotNorm = colors.SymLogNorm(1e-4, linscale=2)
        else:
            plotNorm = colors.Normalize()
    elif par == "imag":
        cmap = plt.get_cmap("RdYlBu")
        zPlot = np.imag(array[flag][indUniFreq])
        if cLevel:
            level = np.concatenate((-np.logspace(0, -4, 33), np.logspace(-4, 0, 33)))
            clevel = np.concatenate((-np.logspace(0, -4, 5), np.logspace(-4, 0, 5)))
        else:
            level = np.linspace(zPlot.min(), zPlot.max(), 100)
            clevel = np.linspace(zPlot.min(), zPlot.max(), 10)
        if colorNorm == "SymLog":
            plotNorm = colors.SymLogNorm(1e-4, linscale=2)
        else:
            plotNorm = colors.Normalize()
    if contour:
        cs = ax.tricontourf(
            x, y, zPlot, levels=level, cmap=cmap, norm=plotNorm
        )  # ,extend='both')
    else:
        uniX, uniY = np.unique(x), np.unique(y)
        X, Y = np.meshgrid(
            np.append(uniX - 25, uniX[-1] + 25), np.append(uniY - 25, uniY[-1] + 25)
        )
        cs = ax.pcolor(
            X,
            Y,
            np.reshape(zPlot, (len(uniY), len(uniX))),
            levels=level,
            cmap=cmap,
            norm=plotNorm,
            edgecolors="k",
            linewidths=0.5,
        )
    if colorbar:
        plt.colorbar(cs, cax=ax.cax, ticks=clevel, format="%1.2e")
    ax.set_title(flag + " " + par, fontsize=8)


def plotIsoStaImpedance(
    ax, loc, array, flag, par="abs", pSym="s", pColor=None, addLabel="", zorder=1
):

    appResFact = 1 / (8 * np.pi ** 2 * 10 ** (-7))
    treshold = 1.0  # 1 meter
    indUniSta = (
        np.sqrt(np.sum((array[["x", "y"]].view((float, 2)) - loc) ** 2, axis=1))
        < treshold
    )
    freq = array["freq"][indUniSta]

    if par == "abs":
        zPlot = np.abs(array[flag][indUniSta])
    elif par == "real":
        zPlot = np.real(array[flag][indUniSta])
    elif par == "imag":
        zPlot = np.imag(array[flag][indUniSta])
    elif par == "res":
        zPlot = (appResFact / freq) * np.abs(array[flag][indUniSta]) ** 2
    elif par == "phs":
        zPlot = np.arctan2(array[flag][indUniSta].imag, array[flag][indUniSta].real) * (
            180 / np.pi
        )

    if not pColor:
        if "xx" in flag:
            lab = "XX"
            pColor = "g"
        elif "xy" in flag:
            lab = "XY"
            pColor = "r"
        elif "yx" in flag:
            lab = "YX"
            pColor = "b"
        elif "yy" in flag:
            lab = "YY"
            pColor = "y"

    ax.plot(
        freq, zPlot, color=pColor, marker=pSym, label=flag + addLabel, zorder=zorder
    )


def plotPsudoSectNSimpedance(
    ax,
    sectDict,
    array,
    flag,
    par="abs",
    colorbar=True,
    colorNorm="None",
    cLevel=None,
    contour=True,
):

    indSect = np.where(sectDict.values()[0] == array[sectDict.keys()[0]])

    # Define the plot axes
    if "x" in sectDict.keys()[0]:
        x = array["y"][indSect]
    else:
        x = array["x"][indSect]
    y = array["freq"][indSect]

    if par == "abs":
        zPlot = np.abs(array[flag][indSect])
        cmap = plt.get_cmap("OrRd_r")  # seismic')
        if cLevel:
            level = np.logspace(0, -5, 31, endpoint=True)
            clevel = np.logspace(0, -4, 5, endpoint=True)
        else:
            level = np.linspace(zPlot.min(), zPlot.max(), 100, endpoint=True)
            clevel = np.linspace(zPlot.min(), zPlot.max(), 10, endpoint=True)

    elif par == "ares":
        zPlot = np.abs(array[flag][indSect]) ** 2 / (
            8 * np.pi ** 2 * 10 ** (-7) * array["freq"][indSect]
        )
        cmap = plt.get_cmap("RdYlBu")  # seismic)
        if cLevel:
            zMax = np.log10(cLevel[1])
            zMin = np.log10(cLevel[0])
        else:
            zMax = np.ceil(np.log10(np.abs(zPlot).max()))
            zMin = np.floor(np.log10(np.abs(zPlot).min()))
        level = np.logspace(zMin, zMax, (zMax - zMin) * 8 + 1, endpoint=True)
        clevel = np.logspace(zMin, zMax, (zMax - zMin) * 2 + 1, endpoint=True)
        plotNorm = colors.LogNorm()

    elif par == "aphs":
        zPlot = np.arctan2(array[flag][indSect].imag, array[flag][indSect].real) * (
            180 / np.pi
        )
        cmap = plt.get_cmap("RdYlBu")  # seismic)
        if cLevel:
            zMax = cLevel[1]
            zMin = cLevel[0]
        else:
            zMax = np.ceil(zPlot).max()
            zMin = np.floor(zPlot).min()
        level = np.arange(zMin, zMax + 0.1, 1)
        clevel = np.arange(zMin, zMax + 0.1, 10)
        plotNorm = colors.Normalize()

    elif par == "real":
        zPlot = np.real(array[flag][indSect])
        cmap = plt.get_cmap("Spectral")  # ('RdYlBu')
        if cLevel:
            zMax = np.log10(cLevel[1])
            zMin = np.log10(cLevel[0])
        else:
            zMax = np.ceil(np.log10(np.abs(zPlot).max()))
            zMin = np.floor(np.log10(np.abs(zPlot).min()))
        level = np.concatenate(
            (
                -np.logspace(zMax, zMin - 0.125, (zMax - zMin) * 8 + 1, endpoint=True),
                np.logspace(zMin - 0.125, zMax, (zMax - zMin) * 8 + 1, endpoint=True),
            )
        )
        clevel = np.concatenate(
            (
                -np.logspace(zMax, zMin, (zMax - zMin) * 1 + 1, endpoint=True),
                np.logspace(zMin, zMax, (zMax - zMin) * 1 + 1, endpoint=True),
            )
        )
        plotNorm = colors.SymLogNorm(np.abs(level).min(), linscale=0.1)
    elif par == "imag":
        zPlot = np.imag(array[flag][indSect])
        cmap = plt.get_cmap("Spectral")  # ('RdYlBu')

        if cLevel:
            zMax = np.log10(cLevel[1])
            zMin = np.log10(cLevel[0])
        else:
            zMax = np.ceil(np.log10(np.abs(zPlot).max()))
            zMin = np.floor(np.log10(np.abs(zPlot).min()))
        level = np.concatenate(
            (
                -np.logspace(zMax, zMin - 0.125, (zMax - zMin) * 8 + 1, endpoint=True),
                np.logspace(zMin - 0.125, zMax, (zMax - zMin) * 8 + 1, endpoint=True),
            )
        )
        clevel = np.concatenate(
            (
                -np.logspace(zMax, zMin, (zMax - zMin) * 1 + 1, endpoint=True),
                np.logspace(zMin, zMax, (zMax - zMin) * 1 + 1, endpoint=True),
            )
        )
        plotNorm = colors.SymLogNorm(np.abs(level).min(), linscale=0.1)

    if colorNorm == "SymLog":
        plotNorm = colors.SymLogNorm(np.abs(level).min(), linscale=0.1)
    elif colorNorm == "Lin":
        plotNorm = colors.Normalize()
    elif colorNorm == "Log":
        plotNorm = colors.LogNorm()
    if contour:
        cs = ax.tricontourf(
            x, y, zPlot, levels=level, cmap=cmap, norm=plotNorm
        )  # ,extend='both')
    else:
        uniX, uniY = np.unique(x), np.unique(y)
        X, Y = np.meshgrid(
            np.append(uniX - 25, uniX[-1] + 25), np.append(uniY - 25, uniY[-1] + 25)
        )
        cs = ax.pcolor(
            X, Y, np.reshape(zPlot, (len(uniY), len(uniX))), cmap=cmap, norm=plotNorm
        )
    if colorbar:
        csB = plt.colorbar(cs, cax=ax.cax, ticks=clevel, format="%1.2e")
        # csB.on_mappable_changed(cs)
        ax.set_title(flag + " " + par, fontsize=8)
        return cs, csB
    return cs, None


def plotPsudoSectNSDiff(
    ax,
    sectDict,
    arrayList,
    flag,
    par="abs",
    colorbar=True,
    colorNorm="SymLog",
    cLevel=None,
    contour=True,
    mask=None,
    useLog=False,
):
    def sortInArr(arr):
        return np.sort(arr, order=["freq", "x", "y", "z"])

    # Find the index for the slice
    indSect0 = np.where(sectDict.values()[0] == arrayList[0][sectDict.keys()[0]])
    indSect1 = np.where(sectDict.values()[0] == arrayList[1][sectDict.keys()[0]])
    # Extract and sort the mats
    arr0 = sortInArr(arrayList[0][indSect0])
    arr1 = sortInArr(arrayList[1][indSect1])

    # Define the plot axes
    if "x" in sectDict.keys()[0]:
        x0 = arr0["y"]
        x1 = arr1["y"]
    else:
        x0 = arr0["x"]
        x1 = arr1["x"]
    y0 = arr0["freq"]
    y1 = arr1["freq"]

    if par == "abs":
        if useLog:
            zPlot = (
                np.log10(np.abs(arr0[flag])) - np.log10(np.abs(arr1[flag]))
            ) / np.log10(np.abs(arr1[flag]))
        else:
            zPlot = (np.abs(arr0[flag]) - np.abs(arr1[flag])) / np.abs(arr1[flag])
        if mask:
            maskInd = np.logical_or(
                np.abs(arr0[flag]) < 1e-3, np.abs(arr1[flag]) < 1e-3
            )
            zPlot = np.ma.array(zPlot)
            zPlot[maskInd] = mask
        cmap = plt.get_cmap("RdYlBu")  # seismic)
    elif par == "ares":
        arF = 1 / (8 * np.pi ** 2 * 10 ** (-7))
        if useLog:
            zPlot = (
                np.log10((arF / arr0["freq"]) * np.abs(arr0[flag]) ** 2)
                - np.log10((arF / arr1["freq"]) * np.abs(arr1[flag]) ** 2)
            ) / np.log10((arF / arr1["freq"]) * np.abs(arr1[flag]) ** 2)
        else:
            zPlot = (
                (arF / arr0["freq"]) * np.abs(arr0[flag]) ** 2
                - (arF / arr1["freq"]) * np.abs(arr1[flag]) ** 2
            ) / ((arF / arr1["freq"]) * np.abs(arr1[flag]) ** 2)
        if mask:
            maskInd = np.logical_or(
                np.abs(arr0[flag]) < 1e-3, np.abs(arr1[flag]) < 1e-3
            )
            zPlot = np.ma.array(zPlot)
            zPlot[maskInd] = mask
        cmap = plt.get_cmap("Spectral")  # seismic)

    elif par == "aphs":
        if useLog:
            zPlot = (
                np.log10(np.arctan2(arr0[flag].imag, arr0[flag].real) * (180 / np.pi))
                - np.log10(np.arctan2(arr1[flag].imag, arr1[flag].real) * (180 / np.pi))
            ) / np.log10(np.arctan2(arr1[flag].imag, arr1[flag].real) * (180 / np.pi))
        else:
            zPlot = (
                np.arctan2(arr0[flag].imag, arr0[flag].real) * (180 / np.pi)
                - np.arctan2(arr1[flag].imag, arr1[flag].real) * (180 / np.pi)
            ) / (np.arctan2(arr1[flag].imag, arr1[flag].real) * (180 / np.pi))
        if mask:
            maskInd = np.logical_or(
                np.abs(arr0[flag]) < 1e-3, np.abs(arr1[flag]) < 1e-3
            )
            zPlot = np.ma.array(zPlot)
            zPlot[maskInd] = mask
        cmap = plt.get_cmap("Spectral")  # seismic)
    elif par == "real":
        if useLog:
            zPlot = (np.log10(arr0[flag].real) - np.log10(arr1[flag].real)) / np.log10(
                arr1[flag].real
            )
        else:
            zPlot = (arr0[flag].real - arr1[flag].real) / arr1[flag].real
        if mask:
            maskInd = np.logical_or(arr0[flag].real < 1e-3, arr1[flag].real < 1e-3)
            zPlot = np.ma.array(zPlot)
            zPlot[maskInd] = mask
        cmap = plt.get_cmap("Spectral")  # ('Spectral')

    elif par == "imag":
        if useLog:
            zPlot = (np.log10(arr0[flag].imag) - np.log10(arr1[flag].imag)) / np.log10(
                arr1[flag].imag
            )
        else:
            zPlot = (arr0[flag].imag - arr1[flag].imag) / arr1[flag].imag
        if mask:
            maskInd = np.logical_or(arr0[flag].imag < 1e-3, arr1[flag].imag < 1e-3)
            zPlot = np.ma.array(zPlot)
            zPlot[maskInd] = mask
        cmap = plt.get_cmap("Spectral")  # ('RdYlBu')

    if cLevel:
        zMax = np.log10(cLevel[1])
        zMin = np.log10(cLevel[0])
    else:
        zMax = np.ceil(np.log10(np.abs(zPlot).max()))
        zMin = np.floor(np.log10(np.abs(zPlot).min()))

    if colorNorm == "SymLog":
        level = np.concatenate(
            (
                -np.logspace(zMax, zMin - 0.125, (zMax - zMin) * 8 + 1, endpoint=True),
                np.logspace(zMin - 0.125, zMax, (zMax - zMin) * 8 + 1, endpoint=True),
            )
        )
        clevel = np.concatenate(
            (
                -np.logspace(zMax, zMin, (zMax - zMin) * 1 + 1, endpoint=True),
                np.logspace(zMin, zMax, (zMax - zMin) * 1 + 1, endpoint=True),
            )
        )
        plotNorm = colors.SymLogNorm(np.abs(level).min(), linscale=0.1)
    elif colorNorm == "Lin":
        if cLevel:
            level = np.arange(
                cLevel[0], cLevel[1] + 0.1, (cLevel[1] - cLevel[0]) / 50.0
            )
            clevel = np.arange(
                cLevel[0], cLevel[1] + 0.1, (cLevel[1] - cLevel[0]) / 10.0
            )
        else:
            level = np.arange(
                zPlot.min(), zPlot.max(), (zPlot.max() - zPlot.min()) / 50.0
            )
            clevel = np.arange(
                zPlot.min(), zPlot.max(), (zPlot.max() - zPlot.min()) / 10.0
            )
        plotNorm = colors.Normalize()
    elif colorNorm == "Log":
        level = np.logspace(zMin - 0.125, zMax, (zMax - zMin) * 8 + 1, endpoint=True)
        clevel = np.logspace(zMin, zMax, (zMax - zMin) * 2 + 1, endpoint=True)
        plotNorm = colors.LogNorm()
    if contour:
        cs = ax.tricontourf(
            x0,
            y0,
            zPlot * 100,
            levels=level * 100,
            cmap=cmap,
            norm=plotNorm,
            extend="both",
        )  # ,extend='both')
    else:
        uniX, uniY = np.unique(x0), np.unique(y0)
        X, Y = np.meshgrid(
            np.append(uniX - 25, uniX[-1] + 25), np.append(uniY - 25, uniY[-1] + 25)
        )
        cs = ax.pcolor(
            X, Y, np.reshape(zPlot, (len(uniY), len(uniX))), cmap=cmap, norm=plotNorm
        )
    if colorbar:
        csB = plt.colorbar(cs, cax=ax.cax, ticks=clevel * 100, format="%1.2e")
        # csB.on_mappable_changed(cs)
        ax.set_title(flag + " " + par + " diff", fontsize=8)
        return cs, csB
    return cs, None



import numpy as np
from scipy.constants import mu_0, pi
from scipy.special import erf


def hzAnalyticDipoleT(r, t, sigma):
    theta = np.sqrt((sigma * mu_0) / (4 * t))
    tr = theta * r
    etr = erf(tr)
    t1 = (9 / (2 * tr ** 2) - 1) * etr
    t2 = (1 / np.sqrt(pi)) * (9 / tr + 4 * tr) * np.exp(-(tr ** 2))
    hz = (t1 - t2) / (4 * pi * r ** 3)
    return hz


def hzAnalyticCentLoopT(a, t, sigma):
    theta = np.sqrt((sigma * mu_0) / (4 * t))
    ta = theta * a
    eta = erf(ta)
    t1 = (3 / (np.sqrt(pi) * ta)) * np.exp(-(ta ** 2))
    t2 = (1 - (3 / (2 * ta ** 2))) * eta
    hz = (t1 + t2) / (2 * a)
    return hz



from .ProblemTDEM import (
    BaseTDEMProblem,
    Problem3D_b,
    Problem3D_e,
    Problem3D_h,
    Problem3D_j,
)
from .FieldsTDEM import FieldsTDEM, Fields3D_b, Fields3D_e, Fields3D_h, Fields3D_j
from .SurveyTDEM import Survey
from . import SrcTDEM as Src
from . import RxTDEM as Rx



from six import string_types
import numpy as np
from . import Utils


class Fields:
    """Fancy Field Storage

    u[:,'phi'] = phi
    print(u[src0,'phi'])

    """

    #: Known fields,   a dict with locations, e.g. {"e": "E", "phi": "CC"}
    knownFields = None
    #: Aliased fields, a dict with [alias, location, function], e.g. {"b":["e","F",lambda(F,e,ind)]}
    aliasFields = None
    #: dtype is the type of the storage matrix. This can be a dictionary.
    dtype = float

    def __init__(self, mesh, survey, **kwargs):
        self.survey = survey
        self.mesh = mesh
        Utils.setKwargs(self, **kwargs)
        self._fields = {}

        if self.knownFields is None:
            raise Exception("knownFields cannot be set to None")
        if self.aliasFields is None:
            self.aliasFields = {}

        allFields = [k for k in self.knownFields] + [a for a in self.aliasFields]
        assert len(allFields) == len(
            set(allFields)
        ), "Aliased fields and Known Fields have overlapping definitions."
        self.startup()

    def startup(self):
        pass

    @property
    def approxSize(self):
        """The approximate cost to storing all of the known fields."""
        sz = 0.0
        for f in self.knownFields:
            loc = self.knownFields[f]
            sz += np.array(self._storageShape(loc)).prod() * 8.0 / (1024 ** 2)
        return f"{sz:e} MB"

    def _storageShape(self, loc):
        nSrc = self.survey.nSrc

        nP = {
            "CC": self.mesh.nC,
            "N": self.mesh.nN,
            "F": self.mesh.nF,
            "E": self.mesh.nE,
        }[loc]

        return (nP, nSrc)

    def _initStore(self, name):
        if name in self._fields:
            return self._fields[name]

        assert name in self.knownFields, "field name is not known."

        loc = self.knownFields[name]

        if type(self.dtype) is dict:
            dtype = self.dtype[name]
        else:
            dtype = self.dtype
        field = np.zeros(self._storageShape(loc), dtype=dtype)

        self._fields[name] = field

        return field

    def _srcIndex(self, srcTestList):
        if type(srcTestList) is slice:
            ind = srcTestList
        else:
            ind = self.survey.getSourceIndex(srcTestList)
        return ind

    def _nameIndex(self, name, accessType):

        if type(name) is slice:
            assert name == slice(
                None, None, None
            ), "Fancy field name slicing is not supported... yet."
            name = None

        if name is None:
            return
        if accessType == "set" and name not in self.knownFields:
            if name in self.aliasFields:
                raise KeyError(
                    "Invalid field name ({!s}) for setter, you can't "
                    "set an aliased property".format(name)
                )
            else:
                raise KeyError(f"Invalid field name ({name!s}) for setter")

        elif accessType == "get" and (
            name not in self.knownFields and name not in self.aliasFields
        ):
            raise KeyError(f"Invalid field name ({name!s}) for getter")
        return name

    def _indexAndNameFromKey(self, key, accessType):
        if type(key) is not tuple:
            key = (key,)
        if len(key) == 1:
            key += (None,)

        assert len(key) == 2, "must be [Src, fieldName]"

        srcTestList, name = key
        name = self._nameIndex(name, accessType)
        ind = self._srcIndex(srcTestList)
        return ind, name

    def __setitem__(self, key, value):
        ind, name = self._indexAndNameFromKey(key, "set")
        if name is None:
            assert (
                type(value) is dict
            ), "New fields must be a dictionary, if field is not specified."
            newFields = value
        elif name in self.knownFields:
            newFields = {name: value}
        else:
            raise Exception("Unknown setter")

        for name in newFields:
            field = self._initStore(name)
            self._setField(field, newFields[name], name, ind)

    def __getitem__(self, key):
        ind, name = self._indexAndNameFromKey(key, "get")
        if name is None:
            out = {}
            for name in self._fields:
                out[name] = self._getField(name, ind)
            return out
        return self._getField(name, ind)

    def _setField(self, field, val, name, ind):
        if isinstance(val, np.ndarray) and (
            field.shape[0] == field.size or val.ndim == 1
        ):
            val = Utils.mkvc(val, 2)
        field[:, ind] = val

    def _getField(self, name, ind):
        if name in self._fields:
            out = self._fields[name][:, ind]
        else:
            # Aliased fields
            alias, loc, func = self.aliasFields[name]

            srcII = np.array(self.survey.srcList)[ind]
            srcII = srcII.tolist()

            if isinstance(func, str):
                assert hasattr(self, func), (
                    "The alias field function is a string, but it does not "
                    "exist in the Fields class."
                )
                func = getattr(self, func)
            out = func(self._fields[alias][:, ind], srcII)
        if out.shape[0] == out.size or out.ndim == 1:
            out = Utils.mkvc(out, 2)
        return out

    def __contains__(self, other):
        if other in self.aliasFields:
            other = self.aliasFields[other][0]
        return self._fields.__contains__(other)


class TimeFields(Fields):
    """Fancy Field Storage for time domain problems

        u[:,'phi', timeInd] = phi
        print(u[src0,'phi'])

    """

    def _storageShape(self, loc):
        nP = {
            "CC": self.mesh.nC,
            "N": self.mesh.nN,
            "F": self.mesh.nF,
            "E": self.mesh.nE,
        }[loc]
        nSrc = self.survey.nSrc
        nT = self.survey.prob.nT + 1
        return (nP, nSrc, nT)

    def _indexAndNameFromKey(self, key, accessType):
        if type(key) is not tuple:
            key = (key,)
        if len(key) == 1:
            key += (None,)
        if len(key) == 2:
            key += (slice(None, None, None),)

        assert len(key) == 3, "must be [Src, fieldName, times]"

        srcTestList, name, timeInd = key

        name = self._nameIndex(name, accessType)
        srcInd = self._srcIndex(srcTestList)

        return (srcInd, timeInd), name

    def _correctShape(self, name, ind, deflate=False):
        srcInd, timeInd = ind
        if name in self.knownFields:
            loc = self.knownFields[name]
        else:
            loc = self.aliasFields[name][1]
        nP, total_nSrc, total_nT = self._storageShape(loc)
        nSrc = np.ones(total_nSrc, dtype=bool)[srcInd].sum()
        nT = np.ones(total_nT, dtype=bool)[timeInd].sum()
        shape = nP, nSrc, nT
        if deflate:
            shape = tuple([s for s in shape if s > 1])
        if len(shape) == 1:
            shape = shape + (1,)
        return shape

    def _setField(self, field, val, name, ind):
        srcInd, timeInd = ind
        shape = self._correctShape(name, ind)
        if isinstance(val, np.ndarray) and val.size == 1:
            val = val[0]
        if np.isscalar(val):
            field[:, srcInd, timeInd] = val
            return
        if val.size != np.array(shape).prod():
            raise ValueError("Incorrect size for data.")
        correctShape = field[:, srcInd, timeInd].shape
        field[:, srcInd, timeInd] = val.reshape(correctShape, order="F")

    def _getField(self, name, ind):
        srcInd, timeInd = ind

        if name in self._fields:
            out = self._fields[name][:, srcInd, timeInd]
        else:
            # Aliased fields
            alias, loc, func = self.aliasFields[name]
            if isinstance(func, str):
                assert hasattr(self, func), (
                    "The alias field function is a string, but it does "
                    "not exist in the Fields class."
                )
                func = getattr(self, func)
            pointerFields = self._fields[alias][:, srcInd, timeInd]
            pointerShape = self._correctShape(alias, ind)
            pointerFields = pointerFields.reshape(pointerShape, order="F")

            timeII = np.arange(self.survey.prob.nT + 1)[timeInd]
            srcII = np.array(self.survey.srcList)[srcInd]
            srcII = srcII.tolist()

            if timeII.size == 1:
                pointerShapeDeflated = self._correctShape(alias, ind, deflate=True)
                pointerFields = pointerFields.reshape(pointerShapeDeflated, order="F")
                out = func(pointerFields, srcII, timeII)
            else:  # loop over the time steps
                nT = pointerShape[2]
                out = list(range(nT))
                for i, TIND_i in enumerate(timeII):
                    fieldI = pointerFields[:, :, i]
                    if fieldI.shape[0] == fieldI.size:
                        fieldI = Utils.mkvc(fieldI, 2)
                    out[i] = func(fieldI, srcII, TIND_i)
                    if out[i].ndim == 1:
                        out[i] = out[i][:, np.newaxis, np.newaxis]
                    elif out[i].ndim == 2:
                        out[i] = out[i][:, :, np.newaxis]
                out = np.concatenate(out, axis=2)

        shape = self._correctShape(name, ind, deflate=True)
        return out.reshape(shape, order="F")



import numpy as np
from scipy.constants import mu_0, epsilon_0

# useful params
def omega(freq):
    """Angular frequency, omega"""
    return 2.0 * np.pi * freq


def k(freq, sigma, mu=mu_0, eps=epsilon_0):
    """ Eq 1.47 - 1.49 in Ward and Hohmann """
    w = omega(freq)
    alp = w * np.sqrt(mu * eps / 2 * (np.sqrt(1.0 + (sigma / (eps * w)) ** 2) + 1))
    beta = w * np.sqrt(mu * eps / 2 * (np.sqrt(1.0 + (sigma / (eps * w)) ** 2) - 1))
    return alp - 1j * beta


def TriangleFun(time, ta, tb):
    """
        Triangular Waveform
        * time: 1D array for time
        * ta: time at peak
        * tb: time at step-off
    """
    out = np.zeros(time.size)
    out[time <= ta] = 1 / ta * time[time <= ta]
    out[(time > ta) & (time < tb)] = (
        -1 / (tb - ta) * (time[(time > ta) & (time < tb)] - tb)
    )
    return out


def TriangleFunDeriv(time, ta, tb):
    """
        Derivative of Triangular Waveform
    """
    out = np.zeros(time.size)
    out[time <= ta] = 1 / ta
    out[(time > ta) & (time < tb)] = -1 / (tb - ta)
    return out


def SineFun(time, ta):
    """
        Sine Waveform
        * time: 1D array for time
        * ta: Pulse Period
    """
    out = np.zeros(time.size)
    out[time <= ta] = np.sin(1.0 / ta * np.pi * time[time <= ta])

    return out


def SineFunDeriv(time, ta):
    """
        Derivative of Sine Waveform
    """
    out = np.zeros(time.size)
    out[time <= ta] = 1.0 / ta * np.pi * np.cos(1.0 / ta * np.pi * time[time <= ta])
    return out


def VTEMFun(time, ta, tb, a):
    """
        VTEM Waveform
        * time: 1D array for time
        * ta: time at peak of exponential part
        * tb: time at step-off
    """
    out = np.zeros(time.size)
    out[time <= ta] = (1 - np.exp(-a * time[time <= ta] / ta)) / (1 - np.exp(-a))
    out[(time > ta) & (time < tb)] = (
        -1 / (tb - ta) * (time[(time > ta) & (time < tb)] - tb)
    )
    return out



import numpy as np
from SimPEG import Utils
from scipy.constants import mu_0, pi
from scipy.special import erf
import matplotlib.pyplot as plt
from .DigFilter import transFiltImpulse, transFilt, setFrequency


def Hzanal(sig, f, r, flag):

    """

        Hz component of analytic solution for half-space (VMD source)
        Src and Rx are on the surface

        .. math::

            H_z  = \\frac{m}{2\\pi k^2 r^5} \
                    \\left( 9 -(9+\\imath\\ kr - 4 k^2r^2 - \\imath k^3r^3)e^{-\\imath kr}\\right)

        * r: Src-Rx offset
        * m: magnetic dipole moment
        * k: propagation constant

        .. math::

            k = \\omega^2\\epsilon\\mu - \\imath\\omega\\mu\\sigma


    """
    mu0 = 4 * np.pi * 1e-7
    w = 2 * np.pi * f
    k = np.sqrt(-1j * w * mu0 * sig)
    Hz = (
        1.0
        / (2 * np.pi * k ** 2 * r ** 5)
        * (
            9
            - (9 + 9 * 1j * k * r - 4 * k ** 2 * r ** 2 - 1j * k ** 3 * r ** 3)
            * np.exp(-1j * k * r)
        )
    )
    if flag == "secondary":
        Hzp = -1 / (4 * np.pi * r ** 3)
        Hz = Hz - Hzp
    return Hz


def HzanalCirc(sig, f, I, a, flag):

    """

        Hz component of analytic solution for half-space (Circular-loop source)
        Src and Rx are on the surface and receiver is located at the center of the loop.

        .. math::

            H_z  = -\\frac{I}{k^2a^3} \
                    \\left( 3 -(3+\\imath\\ ka - k^2a^2 )e^{-\\imath ka}\\right)

        * a: Src-loop radius
        * I: Current intensity

    """
    mu_0 = 4 * np.pi * 1e-7
    w = 2 * np.pi * f
    k = np.sqrt(-1j * w * mu_0 * sig)
    Hz = (
        -I
        / (k ** 2 * a ** 3)
        * (3 - (3 + 3 * 1j * k * a - k ** 2 * a ** 2) * np.exp(-1j * k * a))
    )
    if flag == "secondary":
        Hzp = I / 2.0 / a
        Hz = Hz - Hzp
    return Hz


def dHzdsiganalCirc(sig, f, I, a, flag):

    """
        Compute sensitivity for HzanalCirc by using perturbation

        .. math::

            \\frac{\\partial H_z}{\\partial \\sigma}
            = \\frac{H_z(\\sigma+\\triangle\\sigma)- H_z(\\sigma-\\triangle\\sigma)}
                {2\\triangle\\sigma}
    """
    mu_0 = 4 * np.pi * 1e-7
    w = 2 * np.pi * f
    k = np.sqrt(-1j * w * mu_0 * sig)
    perc = 0.001
    Hzfun = lambda m: HzanalCirc(m, f, I, a, flag)
    dHzdsig = (Hzfun(sig + perc * sig) - Hzfun(sig - perc * sig)) / (2 * perc * sig)
    return dHzdsig


def ColeCole(f, sig_inf=1e-2, eta=0.1, tau=0.1, c=1):
    """
        Computing Cole-Cole model in frequency domain

        .. math::

            \\sigma (\\omega) = \\sigma_{\\infty} -
            \\frac{\\sigma_{\\infty}\\eta}{1+(1-\\eta)(\\imath\\omega\\tau)^c}



        where \\\\(\\\\\\sigma_{\\\\infty}\\\\) is conductivity at infinte frequency,
        \\\\(\\\\\\eta\\\\) is chargeability,
        \\\\(\\\\\\tau\\\\) is chargeability,
        \\\\(\\\\ c\\\\) is chargeability.

    """

    if np.isscalar(sig_inf):
        w = 2 * np.pi * f
        sigma = sig_inf - sig_inf * eta / (1 + (1 - eta) * (1j * w * tau) ** c)
    else:
        sigma = np.zeros((f.size, sig_inf.size), dtype=complex)
        for i in range(f.size):
            w = 2 * np.pi * f[i]
            sigma[i, :] = Utils.mkvc(
                sig_inf - sig_inf * eta / (1 + (1 - eta) * (1j * w * tau) ** c)
            )
    return sigma


def BzAnalT(r, t, sigma):

    theta = np.sqrt((sigma * mu_0) / (4 * t))
    tr = theta * r
    etr = erf(tr)
    t1 = (9 / (2 * tr ** 2) - 1) * etr
    t2 = (1 / np.sqrt(pi)) * (9 / tr + 4 * tr) * np.exp(-(tr ** 2))
    hz = (t1 - t2) / (4 * pi * r ** 3)
    return mu_0 * hz


def BzAnalCircT(a, t, sigma):
    """
        Hz component of analytic solution for half-space (Circular-loop source)
        Src and Rx are on the surface and receiver is located at the center of the loop.
        Src waveform here is step-off.

        .. math::

            h_z  = \\frac{I}{2a} \
                    \\left( \\frac{3}{\\sqrt{\\pi}\\theta a}e^{-\\theta^2a^2}
                    +(1-\\frac{3}{2\\theta^2a^2})erf(\\theta a)\\right)

        .. math::

            \\theta = \\sqrt{\\frac{\\sigma\\mu}{4t}}
    """

    theta = np.sqrt((sigma * mu_0) / (4 * t))
    ta = theta * a
    eta = erf(ta)
    t1 = (3 / (np.sqrt(pi) * ta)) * np.exp(-(ta ** 2))
    t2 = (1 - (3 / (2 * ta ** 2))) * eta
    hz = (t1 + t2) / (2 * a)
    return mu_0 * hz


def dBzdtAnalCircT(a, t, sigma):
    """
        Hz component of analytic solution for half-space (Circular-loop source)
        Src and Rx are on the surface and receiver is located at the center of the loop.
        Src waveform here is step-off.

        .. math::

            \\frac{\\partial h_z}{\\partial t}  = -\\frac{I}{\\mu_0\\sigma a^3} \
                    \\left( 3erf(\\theta a) - \\frac{2}{\\sqrt{\\pi}}\\theta a (3+2\\theta^2 a^2) e^{-\\theta^2a^2}\\right)

        .. math::

            \\theta = \\sqrt{\\frac{\\sigma\\mu}{4t}}
    """
    theta = np.sqrt((sigma * mu_0) / (4 * t))
    const = -1 / (mu_0 * sigma * a ** 3)
    ta = theta * a
    eta = erf(ta)
    t1 = 3 * eta
    t2 = -2 / (np.pi ** 0.5) * ta * (3 + 2 * ta ** 2) * np.exp(-(ta ** 2))
    dhzdt = const * (t1 + t2)
    return mu_0 * dhzdt


def BzAnalCircTCole(a, t, sigma):

    wt, tbase, omega_int = setFrequency(t)
    hz = HzanalCirc(sigma, omega_int / 2 / np.pi, 1.0, a, "secondary")
    # Treatment for inaccuracy in analytic solutions
    ind = omega_int < 0.2
    hz[ind] = 0.0
    hzTD, f0 = transFilt(hz, wt, tbase, omega_int, t)
    return hzTD * mu_0


def dBzdtAnalCircTCole(a, t, sigma):

    wt, tbase, omega_int = setFrequency(t)
    hz = HzanalCirc(sigma, omega_int / 2 / np.pi, 1.0, a, "secondary")
    # Treatment for inaccuracy in analytic solutions
    ind = omega_int < 0.2
    hz[ind] = 0.0
    dhzdtTD = -transFiltImpulse(hz, wt, tbase, omega_int, t)

    return dhzdtTD * mu_0


def diffusion_distance(t, sigma):
    """
        Compute diffusion distance

        .. math::

            d = \\sqrt{\\frac{2t}{\\mu_0\\sigma}}

        assume \\\\(\\\\ \\mu = \\mu_0\\\\) is chargeability
    """

    return np.sqrt(2 * t / mu_0 / sigma)


def skin_depth(f, sigma):
    """
        Compute skin depth

        .. math::

            \\delta = \\sqrt{\\frac{2}{\\omega\\mu_0\\sigma}}

        where \\\\(\\\\ \\omega = 2\\pi f \\\\) is chargeability
    """

    return np.sqrt(2 / mu_0 / sigma / f / (2 * pi))
